from __future__ import print_function
from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

__all__ = ["AbstractViewer"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

import sys

from future.utils import with_metaclass

# Pattern for accessing classmethods on construction
# https://stackoverflow.com/a/13901161/2019542
# adapted for future
# https://python-future.org/compatible_idioms.html#metaclasses
class MetaViewer(type):
    def __new__(mcl, classname, bases, classdict):
        """Create new viewer with class-appropriate doctests"""
        cls = type.__new__(mcl, classname, bases, classdict)

        doc = cls._doctest_body()
        if cls.__doc__ is not None:
            doc = cls.__doc__ + doc
        cls.__doc__ = doc

        return cls

class AbstractViewer(with_metaclass(MetaViewer, object)):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    def __init__(self, vars, title=None, **kwlimits):
        """Create a `AbstractViewer` object.

        Parameters
        ----------
        vars : ~fipy.variables.cellVariable.CellVariable or list
            the `CellVariable` objects to display.
        title : str, optional
            displayed at the top of the `Viewer` window
        xmin, xmax, ymin, ymax, zmin, zmax, datamin, datamax : float, optional
            displayed range of data. Any limit set to
            a (default) value of `None` will autoscale.
        """
        if self.__class__ is AbstractViewer:
            raise NotImplementedError("can't instantiate abstract base class")

        self._vars = self._getSuitableVars(vars)

        self._limits = kwlimits

        if title is None:
            if len(self.vars) == 1:
                title = self.vars[0].name
            else:
                title = ''

        self.title = title

    @property
    def limits(self):
        return self._limits

    @property
    def title(self):
        """The text appearing at the top center.

        (default: if ``len(self.vars) == 1``, the name of
        :attr:`self.vars[0]`, otherwise ``""``.)
        """
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def vars(self):
        """The :class:`~fipy.variables.variable.Variable` or list
        of :class:`~fipy.variables.variable.Variable` objects to
        display."""
        return self._vars

    def _getSuitableVars(self, vars):
        if type(vars) not in [type([]), type(())]:
            vars = [vars]

        return [var for var in vars]

    def setLimits(self, limits={}, **kwlimits):
        """
        Update the limits.

        Parameters
        ----------
        limits : dict, optional
            a (deprecated) alternative to limit keyword arguments
        xmin, xmax, ymin, ymax, zmin, zmax, datamin, datamax : float, optional
            displayed range of data. A 1D `Viewer` will only use `xmin` and
            `xmax`, a 2D viewer will also use `ymin` and `ymax`, and so on. All
            viewers will use `datamin` and `datamax`. Any limit set to a
            (default) value of `None` will autoscale.

        """
        self.limits.update(limits)
        self.limits.update(kwlimits)

    def _getLimit(self, keys, default=None):
        """
        Return the limit associated with the first available key in `keys`

        Parameters
        ----------
        keys : str or :obj:`tuple` of :obj:`str`
            dictionary keys that identify the limits of interest

        Returns
        -------
        float or None
            the value of the limit
        """
        if not (isinstance(keys, list) or isinstance(keys, tuple)):
            keys = (keys,)
        limit = default
        for key in keys:
            limit = self.limits.get(key, limit)
            if limit is not None:
                break

        return limit

    def plot(self, filename=None):
        """
        Update the display of the viewed variables.

        Parameters
        ----------
        filename : str
            If not `None`, the name of a file to save the image into.
        """

        raise NotImplementedError

    def plotMesh(self, filename=None):
        """
        Display a representation of the mesh

        Parameters
        ----------
        filename : str
            If not `None`, the name of a file to save the image into.
        """
        pass

    def _autoscale(self, vars, datamin=None, datamax=None):
        from fipy.tools import numerix

        if datamin is None:
            datamin = 1e300
            for var in vars:
                datamin = min(datamin, numerix.nanmin(var))

        if datamax is None:
            from fipy.tools import numerix
            datamax = -1e300
            for var in vars:
                datamax = max(datamax, numerix.nanmax(var))

        return datamin, datamax

    def _validFileExtensions(self):
        return []

    _saved_stdout = sys.stdout

    @classmethod
    def _doctest_body(cls):
        return ""

    @classmethod
    def _doctest_extra(cls):
        return """
            >>> viewer.title = "{cls.__name__} changed"
            >>> viewer.plot()
            >>> viewer._promptForOpinion()
        """.format(**locals())

    @staticmethod
    def _serial_doctest_raw_input(prompt):
        """Replacement for `raw_input()` that works in doctests
        """
        AbstractViewer._saved_stdout.write("\n")
        AbstractViewer._saved_stdout.write(prompt)
        AbstractViewer._saved_stdout.flush()
        return sys.stdin.readline()

    @staticmethod
    def _doctest_raw_input(prompt):
        """Replacement for `raw_input()` that works in doctests

        This routine attempts to be savvy about running in parallel.
        """
        try:
            from fipy.tools import parallelComm
            parallelComm.Barrier()
            AbstractViewer._saved_stdout.flush()
            if parallelComm.procID == 0:
                txt = AbstractViewer._serial_doctest_raw_input(prompt)
            else:
                txt = ""
            parallelComm.Barrier()
        except ImportError:
            txt = AbstractViewer._serial_doctest_raw_input(prompt)
        return txt


    def _promptForOpinion(self, prompt="Describe any problems with this figure or hit Return: "):
        # This method is usually invoked from a test, which can have a weird
        # state; In particular, it may have a special `raw_input` to prevent user
        # interaction during the test.

        opinion = self._doctest_raw_input(self.__class__.__name__ + ": " + prompt).strip()
        if len(opinion) > 0:
            extensions = ", ".join(self._validFileExtensions())
            if len(extensions) > 0:
                extensions = " (%s)" % extensions
            snapshot = self._doctest_raw_input("Enter a filename%s to save a snapshot (leave blank to skip): " % extensions).strip()
            self.plot(snapshot)
            print(opinion)

    @classmethod
    def _test1D(cls):
        return """
            >>> import fipy as fp
            >>> mesh = fp.Grid1D(nx=100)
            >>> x, = mesh.cellCenters
            >>> xVar = fp.CellVariable(mesh=mesh, name="x", value=x)
            >>> k = fp.Variable(name="k", value=0.)
            >>> viewer = {cls.__name__}(vars=(fp.numerix.sin(k * xVar) + 2,
            ...                               fp.numerix.cos(k * xVar / fp.numerix.pi) + 2),
            ...                 xmin=10, xmax=90,
            ...                 datamin=1.1, datamax=4.0,
            ...                 title="{cls.__name__} test")
            >>> for kval in fp.numerix.arange(0, 0.3, 0.03):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """.format(**locals()) + cls._doctest_extra()

    @classmethod
    def _test2Dbase(cls, mesh):
        return """
            >>> import fipy as fp
            >>> mesh = {mesh}
            >>> x, y = mesh.cellCenters
            >>> xyVar = fp.CellVariable(mesh=mesh, name="x y", value=x * y)
            >>> k = fp.Variable(name="k", value=0.)
            >>> viewer = {cls.__name__}(vars=fp.numerix.sin(k * xyVar) * 1000 + 1002,
            ...                 ymin=0.1, ymax=0.9,
            ...                 # datamin=1.1, datamax=4.0,
            ...                 title="{cls.__name__} test")
            >>> from builtins import range
            >>> for kval in range(10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """.format(**locals()) + cls._doctest_extra()

    @classmethod
    def _test2D(cls):
        return cls._test2Dbase(mesh="fp.Grid2D(nx=50, ny=100, dx=0.1, dy=0.01)")

    @classmethod
    def _test2Dirregular(cls):
        """"""
        return cls._test2Dbase(mesh="""(fp.Grid2D(nx=5, ny=10, dx=0.1, dy=0.1)
            ...         + (fp.Tri2D(nx=5, ny=5, dx=0.1, dy=0.1)
            ...          + ((0.5,), (0.2,))))""")

    @classmethod
    def _test2DvectorBase(cls, mesh):
        return """
            >>> import fipy as fp
            >>> mesh = {mesh}
            >>> x, y = mesh.cellCenters
            >>> xyVar = fp.CellVariable(mesh=mesh, name="x y", value=x * y)
            >>> k = fp.Variable(name="k", value=1.)
            >>> viewer = {cls.__name__}(vars=fp.numerix.sin(k * xyVar).grad,
            ...                 title="{cls.__name__} test")
            >>> for kval in fp.numerix.arange(1, 10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()

            >>> viewer = {cls.__name__}(vars=fp.numerix.sin(k * xyVar).faceGrad,
            ...                 title="{cls.__name__} test")
            >>> from builtins import range
            >>> for kval in range(10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """.format(**locals()) + cls._doctest_extra()

    @classmethod
    def _test2Dvector(cls):
        return cls._test2DvectorBase(mesh="fp.Grid2D(nx=50, ny=100, dx=0.1, dy=0.01)")

    @classmethod
    def _test2DvectorIrregular(cls):
        return cls._test2DvectorBase(mesh="""(fp.Grid2D(nx=5, ny=10, dx=0.1, dy=0.1)
            ...         + (fp.Tri2D(nx=5, ny=5, dx=0.1, dy=0.1)
            ...          + ((0.5,), (0.2,))))""")

    @classmethod
    def _test3D(cls):
        return """
            >>> import fipy as fp
            >>> mesh = fp.Grid3D(nx=50, ny=100, nz=10, dx=0.1, dy=0.01, dz=0.1)
            >>> x, y, z = mesh.cellCenters
            >>> xyzVar = fp.CellVariable(mesh=mesh, name=r"x y z", value=x * y * z)
            >>> k = fp.Variable(name="k", value=0.)
            >>> viewer = {cls.__name__}(vars=fp.numerix.sin(k * xyzVar) + 2,
            ...                     ymin=0.1, ymax=0.9,
            ...                     datamin=1.1, datamax=4.0,
            ...                     title="{cls.__name__} test")
            >>> from builtins import range
            >>> for kval in range(10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """.format(**locals()) + cls._doctest_extra()



from __future__ import division
from __future__ import unicode_literals
from fipy.tools import numerix
from fipy.tools.numerix import MA
from fipy.tools import parallelComm

from fipy.meshes.uniformGrid import UniformGrid
from fipy.meshes.builders import _UniformGrid3DBuilder
from fipy.meshes.builders import _Grid3DBuilder
from fipy.meshes.representations.gridRepresentation import _Grid3DRepresentation
from fipy.meshes.topologies.gridTopology import _Grid3DTopology

__all__ = ["UniformGrid3D"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class UniformGrid3D(UniformGrid):
    """
    3D rectangular-prism Mesh with uniform grid spacing in each dimension.

    X axis runs from left to right.
    Y axis runs from bottom to top.
    Z axis runs from front to back.

    Numbering System:

    Vertices: Numbered in the usual way. X coordinate changes most quickly, then Y, then Z.

    *** arrays are arranged Z, Y, X because in numerix, the final index is the one that changes the most quickly ***

    Cells: Same numbering system as vertices.

    Faces: XY faces numbered first, then XZ faces, then YZ faces. Within each subcategory, it is numbered in the usual way.
    """
    def __init__(self, dx = 1., dy = 1., dz = 1., nx = 1, ny = 1, nz = 1,
                 origin = [[0], [0], [0]], overlap=2, communicator=parallelComm,
                 _RepresentationClass=_Grid3DRepresentation,
                 _TopologyClass=_Grid3DTopology):

        super(UniformGrid3D, self).__init__(communicator=communicator,
                                            _RepresentationClass=_RepresentationClass,
                                            _TopologyClass=_TopologyClass)

        builder = _UniformGrid3DBuilder()

        self.args = {
            'dx': dx,
            'dy': dy,
            'dz': dz,
            'nx': nx,
            'ny': ny,
            'nz': nz,
            'origin': origin,
            'overlap': overlap
        }

        builder.buildGridData([dx, dy, dz], [nx, ny, nz], overlap,
                              communicator, origin)

        ([self.dx, self.dy, self.dz],
         [self.nx, self.ny, self.nz],
         self.dim,
         scale,
         self.globalNumberOfCells,
         self.globalNumberOfFaces,
         self.overlap,
         self.offset,
         self.numberOfVertices,
         self.numberOfFaces,
         self.numberOfCells,
         self.shape,
         self.physicalShape,
         self._meshSpacing,
         self.numberOfXYFaces,
         self.numberOfXZFaces,
         self.numberOfYZFaces,
         self.numberOfHorizontalRows,
         self.numberOfVerticalColumns,
         self.numberOfLayers,
         self.origin) = builder.gridData

    """
    Topology set and calculate
    """

    @property
    def _exteriorFaces(self):
        """
        Return only the faces that have one neighboring cell.
        """
        XYids = self._XYFaceIDs
        XZids = self._XZFaceIDs
        YZids = self._YZFaceIDs

        exteriorIDs = numerix.concatenate((numerix.ravel(XYids[...,      0].swapaxes(0, 1)),
                                           numerix.ravel(XYids[...,     -1].swapaxes(0, 1)),
                                           numerix.ravel(XZids[:,     0,:]),
                                           numerix.ravel(XZids[:,    -1,:]),
                                           numerix.ravel(YZids[ 0,     ...]),
                                           numerix.ravel(YZids[-1,     ...])))

        from fipy.variables.faceVariable import FaceVariable
        exteriorFaces = FaceVariable(mesh=self, value=False)
        exteriorFaces[exteriorIDs] = True
        return exteriorFaces

    @property
    def _interiorFaces(self):
        """
        Return only the faces that have two neighboring cells
        """
        XYids = self._XYFaceIDs
        XZids = self._XZFaceIDs
        YZids = self._YZFaceIDs

        interiorIDs = numerix.concatenate((numerix.ravel(XYids[ ..., 1:-1]),
                                           numerix.ravel(XZids[:, 1:-1,:]),
                                           numerix.ravel(YZids[1:-1,       ...].swapaxes(0, 1))))

        from fipy.variables.faceVariable import FaceVariable
        interiorFaces = FaceVariable(mesh=self, value=False)
        interiorFaces[interiorIDs] = True
        return interiorFaces

    @property
    def _cellToFaceOrientations(self):
        tmp = numerix.take(self.faceCellIDs[0], self.cellFaceIDs)
        return (tmp == MA.indices(tmp.shape)[-1]) * 2 - 1

    @property
    def _adjacentCellIDs(self):
        faceCellIDs = self.faceCellIDs
        return (MA.where(MA.getmaskarray(faceCellIDs[0]), faceCellIDs[1], faceCellIDs[0]).filled(),
                MA.where(MA.getmaskarray(faceCellIDs[1]), faceCellIDs[0], faceCellIDs[1]).filled())

    @property
    def _cellToCellIDs(self):
        ids = MA.zeros((6, self.nx, self.ny, self.nz), 'l')
        indices = numerix.indices((self.nx, self.ny, self.nz))
        nxy = self.nx * self.ny
        same = indices[0] + indices[1] * self.nx + indices[2] * nxy

        ids[0] = same - 1
        ids[1] = same + 1
        ids[2] = same - self.nx
        ids[3] = same + self.nx
        ids[4] = same - nxy
        ids[5] = same + nxy

        if self.nx > 0:
            ids[0, 0,  ...] = MA.masked
            ids[1, -1,  ...] = MA.masked
        if self.ny > 0:
            ids[2,:, 0,:] = MA.masked
            ids[3,:, -1,:] = MA.masked
        if self.nz > 0:
            ids[4, ...,  0] = MA.masked
            ids[5, ..., -1] = MA.masked

        return MA.reshape(ids.swapaxes(1, 3), (6, self.numberOfCells))

    @property
    def _cellToCellIDsFilled(self):
        N = self.numberOfCells
        M = self._maxFacesPerCell
        cellIDs = numerix.repeat(numerix.arange(N)[numerix.newaxis, ...], M, axis=0)
        cellToCellIDs = self._cellToCellIDs
        return MA.where(MA.getmaskarray(cellToCellIDs), cellIDs, cellToCellIDs)

    """
    Geometry set and calculate
    """

    @property
    def _faceAreas(self):
        return numerix.concatenate((numerix.repeat((self.dx * self.dy,), self.numberOfXYFaces),
                                    numerix.repeat((self.dx * self.dz,), self.numberOfXZFaces),
                                    numerix.repeat((self.dy * self.dz,), self.numberOfYZFaces)))

    @property
    def faceNormals(self):
        XYnor = numerix.zeros((3, self.nx, self.ny, self.nz + 1), 'l')
        XYnor[0,     ...] =  1
        XYnor[0,  ..., 0] = -1

        XZnor = numerix.zeros((3, self.nx, self.ny + 1, self.nz), 'l')
        XZnor[1,     ...] =  1
        XZnor[1,:, 0,:] = -1

        YZnor = numerix.zeros((3, self.nx + 1, self.ny, self.nz), 'l')
        YZnor[2,     ...] =  1
        YZnor[2, 0,  ...] = -1

        return numerix.concatenate((numerix.reshape(XYnor[::-1].swapaxes(1, 3), (3, self.numberOfXYFaces)),
                                    numerix.reshape(XZnor[::-1].swapaxes(1, 3), (3, self.numberOfXZFaces)),
                                    numerix.reshape(YZnor[::-1].swapaxes(1, 3), (3, self.numberOfYZFaces))), axis=1)

    @property
    def _cellVolumes(self):
        return numerix.ones(self.numberOfCells, 'd') * self.dx * self.dy * self.dz

    @property
    def _cellCenters(self):
        centers = numerix.zeros((3, self.nx, self.ny, self.nz), 'd')
        indices = numerix.indices((self.nx, self.ny, self.nz))
        centers[0] = (indices[0] + 0.5) * self.dx
        centers[1] = (indices[1] + 0.5) * self.dy
        centers[2] = (indices[2] + 0.5) * self.dz
        ccs = numerix.reshape(centers.swapaxes(1, 3), (3, self.numberOfCells)) + self.origin
        return ccs

    @property
    def _cellDistances(self):
        XYdis = numerix.zeros((self.nz + 1, self.ny, self.nx), 'd')
        XYdis[:] = self.dz
        XYdis[ 0, ...] = self.dz / 2.
        XYdis[-1, ...] = self.dz / 2.

        XZdis = numerix.zeros((self.nz, self.ny + 1, self.nx), 'd')
        XZdis[:] = self.dy
        XZdis[:, 0,:] = self.dy / 2.
        XZdis[:, -1,:] = self.dy / 2.

        YZdis = numerix.zeros((self.nz, self.ny, self.nx + 1), 'd')
        YZdis[:] = self.dx
        YZdis[..., 0] = self.dx / 2.
        YZdis[..., -1] = self.dx / 2.

        return numerix.concatenate((numerix.ravel(XYdis),
                                    numerix.ravel(XZdis),
                                    numerix.ravel(YZdis)))

    @property
    def _faceToCellDistanceRatio(self):
        """how far face is from first to second cell
        
        distance from center of face to center of first cell divided by distance
        between cell centers
        """
        XYdis = numerix.zeros((self.nx, self.ny, self.nz + 1), 'd')
        XYdis[:] = 0.5
        XYdis[..., 0] = 1
        XYdis[..., -1] = 1

        XZdis = numerix.zeros((self.nx, self.ny + 1, self.nz), 'd')
        XZdis[:] = 0.5
        XZdis[:, 0,:] = 1
        XZdis[:, -1,:] = 1

        YZdis = numerix.zeros((self.nx + 1, self.ny, self.nz), 'd')
        YZdis[:] = 0.5
        YZdis[ 0, ...] = 1
        YZdis[-1, ...] = 1

        return numerix.concatenate((numerix.ravel(XYdis.swapaxes(0, 2)),
                                    numerix.ravel(XZdis.swapaxes(0, 2)),
                                    numerix.ravel(YZdis.swapaxes(0, 2))))

    @property
    def _orientedFaceNormals(self):
        return self.faceNormals

    @property
    def _faceTangents1(self):
        XYtan = numerix.zeros((3, self.nx, self.ny, self.nz + 1), 'l')
        XYtan[2, ...] =  1

        XZtan = numerix.zeros((3, self.nx, self.ny + 1, self.nz), 'l')
        XZtan[2, ...] =  1

        YZtan = numerix.zeros((3, self.nx + 1, self.ny, self.nz), 'l')
        YZtan[1, ...] =  1

        return numerix.concatenate((numerix.reshape(XYtan[::-1].swapaxes(1, 3), (3, self.numberOfXYFaces)),
                                    numerix.reshape(XZtan[::-1].swapaxes(1, 3), (3, self.numberOfXZFaces)),
                                    numerix.reshape(YZtan[::-1].swapaxes(1, 3), (3, self.numberOfYZFaces))), axis=1)

    @property
    def _faceTangents2(self):
        XYtan = numerix.zeros((3, self.nx, self.ny, self.nz + 1), 'l')
        XYtan[1, ...] =  1

        XZtan = numerix.zeros((3, self.nx, self.ny + 1, self.nz), 'l')
        XZtan[0, ...] =  1

        YZtan = numerix.zeros((3, self.nx + 1, self.ny, self.nz), 'l')
        YZtan[0, ...] =  1

        return numerix.concatenate((numerix.reshape(XYtan[::-1].swapaxes(1, 3), (3, self.numberOfXYFaces)),
                                    numerix.reshape(XZtan[::-1].swapaxes(1, 3), (3, self.numberOfXZFaces)),
                                    numerix.reshape(YZtan[::-1].swapaxes(1, 3), (3, self.numberOfYZFaces))), axis=1)

    @property
    def _cellToCellDistances(self):
        distances = numerix.zeros((6, self.nx, self.ny, self.nz), 'd')
        distances[0] = self.dx
        distances[1] = self.dx
        distances[2] = self.dy
        distances[3] = self.dy
        distances[4] = self.dz
        distances[5] = self.dz

        distances[0,  0, ...   ] = self.dx / 2.
        distances[1, -1, ...   ] = self.dx / 2.
        distances[2,:,  0,:] = self.dy / 2.
        distances[3,:, -1,:] = self.dy / 2.
        distances[4, ...,     0] = self.dz / 2.
        distances[5, ...,    -1] = self.dz / 2.

        return numerix.reshape(distances.swapaxes(1, 3), (6, self.numberOfCells))

    @property
    def _cellNormals(self):
        normals = numerix.zeros((3, 6, self.numberOfCells), 'd')
        normals[:, 0,:] = [[-1], [ 0], [ 0]]
        normals[:, 1,:] = [[ 1], [ 0], [ 0]]
        normals[:, 2,:] = [[ 0], [-1], [ 0]]
        normals[:, 3,:] = [[ 0], [ 1], [ 0]]
        normals[:, 4,:] = [[ 0], [ 0], [-1]]
        normals[:, 5,:] = [[ 0], [ 0], [ 1]]

        return normals

    @property
    def _cellAreas(self):
        areas = numerix.ones((6, self.numberOfCells), 'd')
        areas[0] = self.dy * self.dz
        areas[1] = self.dy * self.dz
        areas[2] = self.dx * self.dz
        areas[3] = self.dx * self.dz
        areas[4] = self.dx * self.dy
        areas[5] = self.dx * self.dy
        return areas

    @property
    def _cellAreaProjections(self):
        return self._cellAreas * self._cellNormals

##         from numMesh/mesh

    @property
    def _faceCenters(self):

        XYcen = numerix.zeros((3, self.nx, self.ny, self.nz + 1), 'd')
        indices = numerix.indices((self.nx, self.ny, self.nz + 1))
        XYcen[0] = (indices[0] + 0.5) * self.dx
        XYcen[1] = (indices[1] + 0.5) * self.dy
        XYcen[2] = indices[2] * self.dz

        XZcen = numerix.zeros((3, self.nx, self.ny + 1, self.nz), 'd')
        indices = numerix.indices((self.nx, self.ny + 1, self.nz))
        XZcen[0] = (indices[0] + 0.5) * self.dx
        XZcen[1] = indices[1] * self.dy
        XZcen[2] = (indices[2] + 0.5) * self.dz

        YZcen = numerix.zeros((3, self.nx + 1, self.ny, self.nz), 'd')
        indices = numerix.indices((self.nx + 1, self.ny, self.nz))
        YZcen[0] = indices[0] * self.dx
        YZcen[1] = (indices[1] + 0.5) * self.dy
        YZcen[2] = (indices[2] + 0.5) * self.dz

        return numerix.concatenate((numerix.reshape(XYcen.swapaxes(1, 3), (3, self.numberOfXYFaces)),
                                    numerix.reshape(XZcen.swapaxes(1, 3), (3, self.numberOfXZFaces)),
                                    numerix.reshape(YZcen.swapaxes(1, 3), (3, self.numberOfYZFaces))), axis=1) + self.origin

    @property
    def _orientedAreaProjections(self):
        return self._areaProjections

    @property
    def _areaProjections(self):
        return self.faceNormals * self._faceAreas

    @property
    def _faceAspectRatios(self):
        return self._faceAreas / self._cellDistances

    def _translate(self, vector):
        return self.__class__(dx = self.args['dx'], nx = self.args['nx'],
                              dy = self.args['dy'], ny = self.args['ny'],
                              dz = self.args['dz'], nz = self.args['nz'],
                             origin = numerix.array(self.args['origin']) + vector, overlap=self.args['overlap'])

    def __mul__(self, factor):
        return UniformGrid3D(dx = self.dx * factor, nx = self.nx,
                             dy = self.dy * factor, ny = self.ny,
                             dz = self.dz * factor, nz = self.nz,
                             origin = self.origin * factor)

    @property
    def _concatenableMesh(self):
        from fipy.meshes.nonUniformGrid3D import NonUniformGrid3D
        args = self.args.copy()
        origin = args['origin']
        from fipy.tools import serialComm
        args['communicator'] = serialComm
        del args['origin']
        return NonUniformGrid3D(**args) + origin

    @property
    def _cellFaceIDs(self):
        return MA.array(_Grid3DBuilder.createCells(self.nx,
                                                   self.ny,
                                                   self.nz,
                                                   self.numberOfXYFaces,
                                                   self.numberOfXZFaces,
                                                   self.numberOfYZFaces))

    @property
    def _XYFaceIDs(self):
        ids = numerix.arange(0, self.numberOfXYFaces)
        return ids.reshape((self.nz + 1, self.ny, self.nx)).swapaxes(0, 2)

    @property
    def _XZFaceIDs(self):
        ids = numerix.arange(self.numberOfXYFaces, self.numberOfXYFaces + self.numberOfXZFaces)
        return ids.reshape((self.nz, self.ny + 1, self.nx)).swapaxes(0, 2)

    @property
    def _YZFaceIDs(self):
        ids = numerix.arange(self.numberOfXYFaces + self.numberOfXZFaces, self.numberOfFaces)
        return ids.reshape((self.nz, self.ny, self.nx + 1)).swapaxes(0, 2)

    @property
    def _maxFacesPerCell(self):
        return 6

##         from numMesh/mesh

    @property
    def vertexCoords(self):
        return _Grid3DBuilder.createVertices(self.dx, self.dy, self.dz,
                                             self.nx, self.ny, self.nz,
                                             self.numberOfVertices,
                                             self.numberOfHorizontalRows,
                                             self.numberOfVerticalColumns) \
                + self.origin

    @property
    def faceCellIDs(self):
        XYids = MA.zeros((2, self.nx, self.ny, self.nz + 1), 'l')
        indices = numerix.indices((self.nx, self.ny, self.nz + 1))
        XYids[1] = indices[0] + (indices[1] + indices[2] * self.ny) * self.nx
        XYids[0] = XYids[1] - self.nx * self.ny
        XYids[0, ..., 0] = XYids[1, ..., 0]
        XYids[1, ..., 0] = MA.masked
        XYids[1, ..., -1] = MA.masked

        XZids = MA.zeros((2, self.nx, self.ny + 1, self.nz), 'l')
        indices = numerix.indices((self.nx, self.ny + 1, self.nz))
        XZids[1] = indices[0] + (indices[1] + indices[2] * self.ny) * self.nx
        XZids[0] = XZids[1] - self.nx
        XZids[0,:, 0,:] = XZids[1,:, 0,:]
        XZids[1,:, 0,:] = MA.masked
        XZids[1,:, -1,:] = MA.masked

        YZids = MA.zeros((2, self.nx + 1, self.ny, self.nz), 'l')
        indices = numerix.indices((self.nx + 1, self.ny, self.nz))
        YZids[1] = indices[0] + (indices[1] + indices[2] * self.ny) * self.nx
        YZids[0] = YZids[1] - 1
        YZids[0, 0] = YZids[1, 0]
        YZids[1, 0] = MA.masked
        YZids[1, -1] = MA.masked

        return MA.concatenate((XYids.swapaxes(1, 3).reshape((2, self.numberOfXYFaces)),
                               XZids.swapaxes(1, 3).reshape((2, self.numberOfXZFaces)),
                               YZids.swapaxes(1, 3).reshape((2, self.numberOfYZFaces))), axis=1)

##         from common/mesh

    @property
    def _cellVertexIDs(self):
        return self._orderedCellVertexIDs

    @property
    def faceVertexIDs(self):
       return _Grid3DBuilder.createFaces(self.nx, self.ny, self.nz)[1]

    def _calcOrderedCellVertexIDs(self):
        """Correct ordering for `VTK_VOXEL`"""
        ids = numerix.zeros((8, self.nx, self.ny, self.nz), 'l')
        indices = numerix.indices((self.nx, self.ny, self.nz))
        ids[1] = indices[0] + (indices[1] + (indices[2] + 1) * (self.ny + 1) + 1) * (self.nx + 1)
        ids[0] = ids[1] + 1
        ids[3] = indices[0] + (indices[1] + (indices[2] + 1) * (self.ny + 1)) * (self.nx + 1)
        ids[2] = ids[3] + 1
        ids[5] = indices[0] + (indices[1] + indices[2] * (self.ny + 1) + 1) * (self.nx + 1)
        ids[4] = ids[5] + 1
        ids[7] = indices[0] + (indices[1] + indices[2] * (self.ny + 1)) * (self.nx + 1)
        ids[6] = ids[7] + 1

        return numerix.reshape(ids.swapaxes(1, 3), (8, self.numberOfCells))

##     scaling

    def _getNearestCellID(self, points):
        nx = self.args['nx']
        ny = self.args['ny']
        nz = self.args['nz']

        x0, y0, z0 = self.cellCenters[..., 0]
        xi, yi, zi = points
        nx, ny, nz = self.shape
        dx, dy, dz = self.dx, self.dy, self.dz

        i = numerix.array(numerix.rint(((xi - x0) / dx)), 'l')
        i[i < 0] = 0
        i[i > nx - 1] = nx - 1

        j = numerix.array(numerix.rint(((yi - y0) / dy)), 'l')
        j[j < 0] = 0
        j[j > ny - 1]  = ny - 1

        k = numerix.array(numerix.rint(((zi - z0) / dz)), 'l')
        k[k < 0] = 0
        k[k > nz - 1]  = nz - 1

        return k * ny * nx + j * nx + i

    def _test(self):
        """
        These tests are not useful as documentation, but are here to ensure
        everything works as expected.

            >>> dx = 0.5
            >>> dy = 2.
            >>> dz = 4.
            >>> nx = 3
            >>> ny = 2
            >>> nz = 1

            >>> mesh = UniformGrid3D(nx=nx, ny=ny, nz=nz, dx=dx, dy=dy, dz=dz)

            >>> XYFaceIDs = numerix.array((((0, 6), (3, 9)),
            ...                            ((1, 7), (4, 10)),
            ...                            ((2, 8), (5, 11))))
            >>> print(numerix.allequal(XYFaceIDs, mesh._XYFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> XZFaceIDs = numerix.array((((12,), (15,), (18,)),
            ...                            ((13,), (16,), (19,)),
            ...                            ((14,), (17,), (20,))))
            >>> print(numerix.allequal(mesh._XZFaceIDs, XZFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> YZFaceIDs = numerix.array((((21,), (25,)),
            ...                            ((22,), (26,)),
            ...                            ((23,), (27,)),
            ...                            ((24,), (28,))))
            >>> print(numerix.allequal(mesh._YZFaceIDs, YZFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> adjacentCellIDs = (numerix.array([0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2, 0, 1, 2, 3, 4, 5, 0, 0, 1, 2, 3, 3, 4, 5]),
            ...                    numerix.array([0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 3, 4, 5, 0, 1, 2, 2, 3, 4, 5, 5]))
            >>> print(numerix.allequal(mesh._adjacentCellIDs, adjacentCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> vertices = numerix.array(((0., 1., 2., 3., 0., 1., 2., 3., 0., 1., 2., 3., 0., 1., 2., 3., 0., 1., 2., 3., 0., 1., 2., 3.),
            ...                           (0., 0., 0., 0., 1., 1., 1., 1., 2., 2., 2., 2., 0., 0., 0., 0., 1., 1., 1., 1., 2., 2., 2., 2.),
            ...                           (0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.)))
            >>> vertices *= numerix.array([[dx], [dy], [dz]])

            >>> print(numerix.allequal(vertices, mesh.vertexCoords)) # doctest: +PROCESSOR_0
            True

            >>> faces = numerix.array((( 0,  1,  2,  4,  5,  6, 12, 13, 14, 16, 17, 18,  0,  1,  2,  4,  5,  6,  8,  9, 10,  0,  1,  2,  3,  4,  5,  6,  7),
            ...                        ( 1,  2,  3,  5,  6,  7, 13, 14, 15, 17, 18, 19,  1,  2,  3,  5,  6,  7,  9, 10, 11,  4,  5,  6,  7,  8,  9, 10, 11),
            ...                        ( 5,  6,  7,  9, 10, 11, 17, 18, 19, 21, 22, 23, 13, 14, 15, 17, 18, 19, 21, 22, 23, 16, 17, 18, 19, 20, 21, 22, 23),
            ...                        ( 4,  5,  6,  8,  9, 10, 16, 17, 18, 20, 21, 22, 12, 13, 14, 16, 17, 18, 20, 21, 22, 12, 13, 14, 15, 16, 17, 18, 19)))
            >>> print(numerix.allclose(faces, mesh.faceVertexIDs)) # doctest: +PROCESSOR_0
            True

            >>> cells = numerix.array(((21, 22, 23, 25, 26, 27),
            ...                        (22, 23, 24, 26, 27, 28),
            ...                        (12, 13, 14, 15, 16, 17),
            ...                        (15, 16, 17, 18, 19, 20),
            ...                        ( 0,  1,  2,  3,  4,  5),
            ...                        ( 6,  7,  8,  9, 10, 11)))
            >>> print(numerix.allequal(cells, mesh.cellFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> externalFaces = numerix.array((0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 18, 19, 20, 21, 24, 25, 28))
            >>> print(numerix.allequal(externalFaces,
            ...                        numerix.nonzero(mesh.exteriorFaces))) # doctest: +PROCESSOR_0
            True

            >>> internalFaces = numerix.array((15, 16, 17, 22, 23, 26, 27))
            >>> print(numerix.allequal(internalFaces,
            ...                        numerix.nonzero(mesh.interiorFaces))) # doctest: +PROCESSOR_0
            True

            >>> from fipy.tools.numerix import MA
            >>> faceCellIds = MA.masked_values((( 0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2, 0, 1, 2, 3, 4, 5, 0, 0, 1, 2, 3, 3, 4, 5),
            ...                                 (-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 3, 4, 5, -1, -1, -1, -1, 1, 2, -1, -1, 4, 5, -1)), -1)
            >>> print(numerix.allequal(faceCellIds, mesh.faceCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> xy = dx * dy
            >>> xz = dx * dz
            >>> yz = dy * dz
            >>> faceAreas = numerix.array((xy, xy, xy, xy, xy, xy, xy, xy, xy, xy, xy, xy,
            ...                            xz, xz, xz, xz, xz, xz, xz, xz, xz,
            ...                            yz, yz, yz, yz, yz, yz, yz, yz))
            >>> print(numerix.allclose(faceAreas, mesh._faceAreas, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            1

            >>> faceCoords = numerix.take(vertices, faces, axis=1)
            >>> faceCenters = (faceCoords[..., 0,:] + faceCoords[..., 1,:] + faceCoords[..., 2,:] + faceCoords[..., 3,:]) / 4.
            >>> print(numerix.allclose(faceCenters, mesh.faceCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> faceNormals = numerix.array((( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 1, 1, 1, -1, 1, 1, 1),
            ...                              ( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0),
            ...                              (-1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)))
            >>> print(numerix.allclose(faceNormals, mesh.faceNormals, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellToFaceOrientations = numerix.array((( 1, -1, -1, 1, -1, -1),
            ...                                         ( 1, 1, 1, 1, 1, 1),
            ...                                         ( 1, 1, 1, -1, -1, -1),
            ...                                         ( 1, 1, 1, 1, 1, 1),
            ...                                         ( 1, 1, 1, 1, 1, 1),
            ...                                         ( 1, 1, 1, 1, 1, 1)))
            >>> print(numerix.allequal(cellToFaceOrientations, mesh._cellToFaceOrientations)) # doctest: +PROCESSOR_0
            True

            >>> cellVolumes = numerix.array((dx*dy*dz, dx*dy*dz, dx*dy*dz, dx*dy*dz, dx*dy*dz, dx*dy*dz))
            >>> print(numerix.allclose(cellVolumes, mesh.cellVolumes, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellCenters = numerix.array(((   dx/2., 3.*dx/2., 5.*dx/2.,   dx/2., 3.*dx/2., 5.*dx/2.),
            ...                              (   dy/2.,    dy/2.,    dy/2., 3.*dy/2., 3.*dy/2., 3.*dy/2.),
            ...                              (   dz/2.,    dz/2.,    dz/2.,   dz/2.,    dz/2.,    dz/2.)))
            >>> print(numerix.allclose(cellCenters, mesh.cellCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> cellDistances = numerix.array((dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2,
            ...                                dy/2, dy/2, dy/2, dy, dy, dy, dy/2, dy/2, dy/2,
            ...                                dx/2, dx, dx, dx/2, dx/2, dx, dx, dx/2))
            >>> print(numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> faceToCellDistances = MA.masked_values(((dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2),
            ...                                         (  -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, dy/2, dy/2, dy/2,   -1,   -1,   -1,   -1, dx/2, dx/2,   -1,   -1, dx/2, dx/2,   -1)), -1)
            >>> faceToCellDistanceRatios = faceToCellDistances[0] / cellDistances
            >>> print(numerix.allclose(faceToCellDistanceRatios, mesh._faceToCellDistanceRatio, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> areaProjections = faceNormals * faceAreas
            >>> print(numerix.allclose(areaProjections, mesh._areaProjections, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> tangents1 = numerix.array(((1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0),
            ...                            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1),
            ...                            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)))
            >>> print(numerix.allclose(tangents1, mesh._faceTangents1, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> tangents2 = numerix.array(((0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            ...                            (1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            ...                            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)))
            >>> print(numerix.allclose(tangents2, mesh._faceTangents2, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellIDs = MA.masked_values(((-1, 0, 1, -1, 3, 4),
            ...                                   (1, 2, -1, 4, 5, -1),
            ...                                   (-1, -1, -1, 0, 1, 2),
            ...                                   (3, 4, 5, -1, -1, -1),
            ...                                   (-1, -1, -1, -1, -1, -1),
            ...                                   (-1, -1, -1, -1, -1, -1)), -1)
            >>> print(numerix.allequal(mesh._cellToCellIDs, cellToCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellIDsFilled = numerix.array([[0, 0, 1, 3, 3, 4],
            ...                                      [1, 2, 2, 4, 5, 5],
            ...                                      [0, 1, 2, 0, 1, 2],
            ...                                      [3, 4, 5, 3, 4, 5],
            ...                                      [0, 1, 2, 3, 4, 5],
            ...                                      [0, 1, 2, 3, 4, 5]])
            >>> print(numerix.allequal(mesh._cellToCellIDsFilled, cellToCellIDsFilled)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellDistances = numerix.take(cellDistances, cells)
            >>> print(numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellNormals = numerix.array((((-1, -1, -1, -1, -1, -1),
            ...                               ( 1,  1,  1,  1,  1,  1),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0)),
            ...                              (( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               (-1, -1, -1, -1, -1, -1),
            ...                               ( 1,  1,  1,  1,  1,  1),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0)),
            ...                              (( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               (-1, -1, -1, -1, -1, -1),
            ...                               ( 1,  1,  1,  1,  1,  1))))
            >>> print(numerix.allclose(cellNormals, mesh._cellNormals, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellAreaProjections = numerix.array((((-yz, -yz, -yz, -yz, -yz, -yz),
            ...                                       ( yz, yz, yz, yz, yz, yz),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0)),
            ...                                      ((  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (-xz, -xz, -xz, -xz, -xz, -xz),
            ...                                       ( xz, xz, xz, xz, xz, xz),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0)),
            ...                                      ((  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (  0,  0,  0,  0,  0,  0),
            ...                                       (-xy, -xy, -xy, -xy, -xy, -xy),
            ...                                       ( xy, xy, xy, xy, xy, xy))))
            >>> print(numerix.allclose(cellAreaProjections, mesh._cellAreaProjections, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellVertexIDs = numerix.array((17, 16, 13, 12, 5, 4, 1, 0))
            >>> cellVertexIDs = numerix.array((cellVertexIDs, cellVertexIDs + 1, cellVertexIDs + 2,
            ...                                cellVertexIDs + 4, cellVertexIDs + 5, cellVertexIDs + 6))
            >>> cellVertexIDs = cellVertexIDs.swapaxes(0, 1)
            >>> print(numerix.allclose(mesh._cellVertexIDs, cellVertexIDs)) # doctest: +PROCESSOR_0
            True

            >>> from fipy.tools import dump
            >>> (f, filename) = dump.write(mesh, extension = '.gz')
            >>> unpickledMesh = dump.read(filename, f)

            >>> print(numerix.allequal(mesh.cellCenters, unpickledMesh.cellCenters))
            True

            # Bug #130 & #135 and issue #470 are because we only checked a mesh with `nz` of 1

            >>> nx = 1
            >>> ny = 2
            >>> nz = 3

            >>> mesh = UniformGrid3D(nx=nx, ny=ny, nz=nz, dx=dx, dy=dy, dz=dz)

            >>> cellVertexIDs = numerix.array((9, 8, 7, 6, 3, 2, 1, 0))
            >>> cellVertexIDs = numerix.array((cellVertexIDs, cellVertexIDs + 2, cellVertexIDs + 6,
            ...                                cellVertexIDs + 8, cellVertexIDs + 12, cellVertexIDs + 14))
            >>> cellVertexIDs = cellVertexIDs.swapaxes(0, 1)
            >>> print(numerix.allclose(mesh._cellVertexIDs, cellVertexIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellIDs = MA.masked_values(((-1, -1, -1, -1, -1, -1),
            ...                                   (-1, -1, -1, -1, -1, -1),
            ...                                   (-1, 0, -1, 2, -1, 4),
            ...                                   ( 1, -1, 3, -1, 5, -1),
            ...                                   (-1, -1, 0, 1, 2, 3),
            ...                                   ( 2, 3, 4, 5, -1, -1)), -1)
            >>> print(numerix.allequal(mesh._cellToCellIDs, cellToCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellDistances = numerix.array((dz/2, dz/2, dz, dz, dz, dz, dz/2, dz/2,
            ...                                dy/2, dy, dy/2, dy/2, dy, dy/2, dy/2, dy, dy/2,
            ...                                dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, ))
            >>> print(numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> faceToCellDistances = MA.masked_values(((dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2),
            ...                                         (  -1,   -1, dz/2, dz/2, dz/2, dz/2,   -1,   -1,   -1, dy/2,   -1,   -1, dy/2,   -1,   -1, dy/2,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1)), -1)
            >>> faceToCellDistanceRatios = faceToCellDistances[0] / cellDistances
            >>> print(numerix.allclose(faceToCellDistanceRatios, mesh._faceToCellDistanceRatio, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cells = numerix.array(((17, 19, 21, 23, 25, 27),
            ...                        (18, 20, 22, 24, 26, 28),
            ...                        ( 8,  9, 11, 12, 14, 15),
            ...                        ( 9, 10, 12, 13, 15, 16),
            ...                        ( 0,  1,  2,  3,  4,  5),
            ...                        ( 2,  3,  4,  5,  6,  7)))
            >>> print(numerix.allequal(cells, mesh.cellFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellDistances = numerix.take(cellDistances, cells)
            >>> print(numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            
            >>> nx = 3
            >>> ny = 1
            >>> nz = 2

            >>> mesh = UniformGrid3D(nx=nx, ny=ny, nz=nz, dx=dx, dy=dy, dz=dz)

            >>> cellVertexIDs = numerix.array((13, 12, 9, 8, 5, 4, 1, 0))
            >>> cellVertexIDs = numerix.array((cellVertexIDs, cellVertexIDs + 1, cellVertexIDs + 2,
            ...                                cellVertexIDs + 8, cellVertexIDs + 9, cellVertexIDs + 10))
            >>> cellVertexIDs = cellVertexIDs.swapaxes(0, 1)
            >>> print(numerix.allclose(mesh._cellVertexIDs, cellVertexIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellIDs = MA.masked_values(((-1, 0, 1, -1, 3, 4),
            ...                                   ( 1, 2, -1, 4, 5, -1),
            ...                                   (-1, -1, -1, -1, -1, -1),
            ...                                   (-1, -1, -1, -1, -1, -1),
            ...                                   (-1, -1, -1, 0, 1, 2),
            ...                                   ( 3, 4, 5, -1, -1, -1)), -1)
            >>> print(numerix.allequal(mesh._cellToCellIDs, cellToCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellDistances = numerix.array((dz/2, dz/2, dz/2, dz, dz, dz, dz/2, dz/2, dz/2,
            ...                                dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2,
            ...                                dx/2, dx, dx, dx/2, dx/2, dx, dx, dx/2))
            >>> print(numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> faceToCellDistances = MA.masked_values(((dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2, dx/2),
            ...                                         (  -1,   -1,   -1, dz/2, dz/2, dz/2,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, dx/2, dx/2,   -1,   -1, dx/2, dx/2,   -1)), -1)
            >>> faceToCellDistanceRatios = faceToCellDistances[0] / cellDistances
            >>> print(numerix.allclose(faceToCellDistanceRatios, mesh._faceToCellDistanceRatio, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cells = numerix.array(((21, 22, 23, 25, 26, 27),
            ...                        (22, 23, 24, 26, 27, 28),
            ...                        ( 9, 10, 11, 15, 16, 17),
            ...                        (12, 13, 14, 18, 19, 20),
            ...                        ( 0,  1,  2,  3,  4,  5),
            ...                        ( 3,  4,  5,  6,  7,  8)))
            >>> print(numerix.allequal(cells, mesh.cellFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellDistances = numerix.take(cellDistances, cells)
            >>> print(numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True


            # issue 470
            # 3x2x1 mesh above did not exhibit the failure of
            # `_cellToCellIDs` and `_cellToCellDistances` to present proper results

            >>> dx = 0.5
            >>> dy = 2.
            >>> dz = 4.
            >>> nx = 4
            >>> ny = 3
            >>> nz = 2

            >>> mesh = UniformGrid3D(nx=nx, ny=ny, nz=nz, dx=dx, dy=dy, dz=dz)

            >>> XYFaceIDs = numerix.array((((0, 12, 24), (4, 16, 28), ( 8, 20, 32)),
            ...                            ((1, 13, 25), (5, 17, 29), ( 9, 21, 33)),
            ...                            ((2, 14, 26), (6, 18, 30), (10, 22, 34)),
            ...                            ((3, 15, 27), (7, 19, 31), (11, 23, 35))))
            >>> print(numerix.allequal(XYFaceIDs, mesh._XYFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> XZFaceIDs = numerix.array((((36, 52), (40, 56), (44, 60), (48, 64)),
            ...                            ((37, 53), (41, 57), (45, 61), (49, 65)),
            ...                            ((38, 54), (42, 58), (46, 62), (50, 66)),
            ...                            ((39, 55), (43, 59), (47, 63), (51, 67))))
            >>> print(numerix.allequal(mesh._XZFaceIDs, XZFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> YZFaceIDs = numerix.array((((68, 83), (73, 88), (78, 93)),
            ...                            ((69, 84), (74, 89), (79, 94)),
            ...                            ((70, 85), (75, 90), (80, 95)),
            ...                            ((71, 86), (76, 91), (81, 96)),
            ...                            ((72, 87), (77, 92), (82, 97))))
            >>> print(numerix.allequal(mesh._YZFaceIDs, YZFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> adjacentCellIDs = (numerix.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 
            ...                                   0, 1, 2, 3, 0, 1, 2, 3, 4, 5,  6,  7, 8, 9, 10, 11, 12, 13, 14, 15, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
            ...                                   0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 8, 8,  9, 10, 11, 12, 12, 13, 14, 15, 16, 16, 17, 18, 19, 20, 20, 21, 22, 23]), 
            ...                    numerix.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
            ...                                   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 20, 21, 22, 23,
            ...                                   0, 1, 2, 3, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 13, 14, 15, 15, 16, 17, 18, 19, 19, 20, 21, 22, 23, 23]))
            >>> print(numerix.allequal(mesh._adjacentCellIDs, adjacentCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> vertices = numerix.array(((0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4.),
            ...                           (0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 2., 2., 2., 2., 2., 3., 3., 3., 3., 3., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 2., 2., 2., 2., 2., 3., 3., 3., 3., 3., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 2., 2., 2., 2., 2., 3., 3., 3., 3., 3.),
            ...                           (0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2.)))
            >>> vertices *= numerix.array([[dx], [dy], [dz]])

            >>> print(numerix.allequal(vertices, mesh.vertexCoords)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellIDs = MA.masked_values(((-1, 0, 1, 2, -1, 4, 5, 6, -1, 8, 9, 10, -1, 12, 13, 14, -1, 16, 17, 18, -1, 20, 21, 22),
            ...                                   ( 1, 2, 3, -1, 5, 6, 7, -1, 9, 10, 11, -1, 13, 14, 15, -1, 17, 18, 19, -1, 21, 22, 23, -1),
            ...                                   (-1, -1, -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, -1, -1, -1, -1, 12, 13, 14, 15, 16, 17, 18, 19),
            ...                                   ( 4, 5, 6, 7, 8, 9, 10, 11, -1, -1, -1, -1, 16, 17, 18, 19, 20, 21, 22, 23, -1, -1, -1, -1),
            ...                                   (-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11),
            ...                                   (12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1)), -1)
            >>> print(numerix.allequal(mesh._cellToCellIDs, cellToCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellDistances = numerix.array((dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz, dz, dz, dz, dz, dz, dz, dz, dz, dz, dz, dz, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2, dz/2,
            ...                                dy/2, dy/2, dy/2, dy/2, dy, dy, dy, dy, dy, dy, dy, dy, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy/2, dy, dy, dy, dy, dy, dy, dy, dy, dy/2, dy/2, dy/2, dy/2,
            ...                                dx/2, dx, dx, dx, dx/2, dx/2, dx, dx, dx, dx/2, dx/2, dx, dx, dx, dx/2, dx/2, dx, dx, dx, dx/2, dx/2, dx, dx, dx, dx/2, dx/2, dx, dx, dx, dx/2))
            >>> print(numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cells = numerix.array(((68, 69, 70, 71, 73, 74, 75, 76, 78, 79, 80, 81, 83, 84, 85, 86, 88, 89, 90, 91, 93, 94, 95, 96),
            ...                        (69, 70, 71, 72, 74, 75, 76, 77, 79, 80, 81, 82, 84, 85, 86, 87, 89, 90, 91, 92, 94, 95, 96, 97),
            ...                        (36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63),
            ...                        (40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67),
            ...                        ( 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23),
            ...                        (12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35)))
            >>> print(numerix.allequal(cells, mesh.cellFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellDistances = numerix.take(cellDistances, cells)
            >>> print(numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            Oh, how boring. We'll assume the 3x2x1 permutations are good enough for everything else until proven otherwise.
            
        Ensure that ghost faces are excluded from accumulating operations
        (#856).  Four exterior surfaces of :math:`10\times 10\times 10`
        cube mesh should each have a total area of 100, regardless of
        partitioning.

            >>> cube = UniformGrid3D(nx=10, dx=1., ny=10, dy=1., nz=10, dz=1.)

            >>> area = (cube._faceAreas * cube.facesBottom).sum()
            >>> print(numerix.allclose(area, 100))
            True

            >>> area = (cube._faceAreas * cube.facesTop).sum()
            >>> print(numerix.allclose(area, 100))
            True

            >>> area = (cube._faceAreas * cube.facesLeft).sum()
            >>> print(numerix.allclose(area, 100))
            True

            >>> area = (cube._faceAreas * cube.facesRight).sum()
            >>> print(numerix.allclose(area, 100))
            True

            >>> area = (cube._faceAreas * cube.facesFront).sum()
            >>> print(numerix.allclose(area, 100))
            True

            >>> area = (cube._faceAreas * cube.facesBack).sum()
            >>> print(numerix.allclose(area, 100))
            True

        Ensure no double-counting of faces on boundary between partitions.

            >>> area = (cube._faceAreas * (cube.faceCenters[2] == 5.)).sum()
            >>> print(numerix.allclose(area, 100)) # doctest: +PARALLEL_2
            True

        Size of global value should not depend on number of processors (#400)

            >>> print(cube.cellCenters.globalValue.shape)
            (3, 1000)
            >>> print(cube.faceCenters.globalValue.shape)
            (3, 3300)
        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.explicitUpwindConvectionTerm import ExplicitUpwindConvectionTerm
from fipy.tools import numerix

__all__ = ["VanLeerConvectionTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class VanLeerConvectionTerm(ExplicitUpwindConvectionTerm):

    def _getGradient(self, normalGradient, gradUpwind):
        gradUpUpwind = -gradUpwind + 2 * normalGradient

        avg = 0.5 * (abs(gradUpwind) + abs(gradUpUpwind))
        min3 = numerix.minimum(numerix.minimum(abs(2 * gradUpwind),
                                               abs(2 * gradUpUpwind)), avg)

        grad = numerix.where(gradUpwind * gradUpUpwind < 0.,
                             0.,
                             numerix.where(gradUpUpwind > 0.,
                                           min3,
                                           -min3))

        return grad

    def _getOldAdjacentValues(self, oldArray, id1, id2, dt):
        oldArray1, oldArray2 = ExplicitUpwindConvectionTerm._getOldAdjacentValues(self, oldArray, id1, id2, dt)

        mesh = oldArray.mesh

        interiorIDs = numerix.nonzero(mesh.interiorFaces)[0]
        interiorFaceAreas = numerix.take(mesh._faceAreas, interiorIDs)
        interiorFaceNormals = numerix.take(mesh._orientedFaceNormals, interiorIDs, axis=-1)

        # Courant-Friedrichs-Levy number
        interiorCFL = abs(numerix.take(self._getGeomCoeff(oldArray), interiorIDs)) * dt

        gradUpwind = (oldArray2 - oldArray1) / numerix.take(mesh._cellDistances, interiorIDs)

        vol1 = numerix.take(mesh.cellVolumes, id1)

        oldArray1 += 0.5 * self._getGradient(numerix.dot(numerix.take(oldArray.grad, id1, axis=-1), interiorFaceNormals), gradUpwind) \
            * (vol1 - interiorCFL) / interiorFaceAreas

        vol2 = numerix.take(mesh.cellVolumes, id2)

        oldArray2 += 0.5 * self._getGradient(numerix.dot(numerix.take(oldArray.grad, id2, axis=-1), -interiorFaceNormals), -gradUpwind) \
            * (vol2 - interiorCFL) / interiorFaceAreas

        return oldArray1, oldArray2

    def _test(self):
        """
        Test for ticket:441.

        >>> from fipy import *
        >>> m = Grid1D()
        >>> c = CellVariable(mesh=m)
        >>> e = VanLeerConvectionTerm(((1,),))
        >>> e.solve(c) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
           ...
        TransientTermError: The equation requires a TransientTerm with explicit convection.

        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from PyTrilinos import Epetra
from PyTrilinos import EpetraExt
from PyTrilinos import Amesos
from PyTrilinos import AztecOO
from PyTrilinos import ML
from PyTrilinos import IFPACK

from fipy import input
from fipy.solvers.trilinos.trilinosSolver import TrilinosSolver
from fipy.tools import numerix

__all__ = ["TrilinosMLTest"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class TrilinosMLTest(TrilinosSolver):

    """
    This solver class does not actually solve the system, but outputs
    information about what ML preconditioner settings will work best.
    """

    def __init__(self, tolerance=1e-10, iterations=5, MLOptions={}, testUnsupported = False):
        """
        For detailed information on the possible parameters for ML, see
        http://trilinos.sandia.gov/packages/ml/documentation.html

        Currently, passing options to Aztec through ML is not supported.

        Parameters
        ----------
        tolerance : float
            Required error tolerance.
        iterations : int
            Maximum number of iterative steps to perform.
        MLOptions : dict
            Options to pass to ML.
            This will be passed to `ML.SetParameterList`.
        testUnsupported : bool
            Test smoothers that are not currently implemented in
            preconditioner objects.
        """

        TrilinosSolver.__init__(self, tolerance=tolerance,
                                iterations=iterations)

        self.MLOptions = MLOptions
        if "output" not in self.MLOptions:
            self.MLOptions["output"] = 0

        if "test: max iters" not in self.MLOptions:
            self.MLOptions["test: max iters"] = iterations

        if "test: tolerance" not in self.MLOptions:
            self.MLOptions["test: tolerance"] = tolerance


        unsupportedSmoothers = ["Jacobi", "Gauss-Seidel", "block Gauss-Seidel", "ParaSails", "IFPACK", "ML"]

        if not testUnsupported:
            for smoother in unsupportedSmoothers:
                if ("test: " + smoother) not in self.MLOptions:
                    self.MLOptions["test: " + smoother] = False



    def _applyTrilinosSolver(self, A, LHS, RHS):

        Prec = ML.MultiLevelPreconditioner(A, False)

        Prec.SetParameterList(self.MLOptions)
        Prec.ComputePreconditioner()

        Prec.TestSmoothers()
        input("Results of preconditioner tests shown above. Currently, the first tests in the 'Gauss-Seidel (sym)','Aztec preconditioner', and 'Aztec as solver' sections indicate the expected performance of the MultilevelSGSPreconditioner, MultilevelDDPreconditioner, and MultilevelSolverSmootherPreconditioner classes, respectively.\n\nPress enter to quit.")
        import sys
        sys.exit(0)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

def _UnaryOperatorVariable(operatorClass=None):
    """
    Test `binOp` pickling

    >>> from fipy import Grid1D, FaceVariable, CellVariable, dump, Variable
    >>> import os, sys
    >>> m = Grid1D()
    >>> vs = (CellVariable(mesh=m, value=2.), FaceVariable(mesh=m, value=3.), Variable(4))
    >>> tmp = []
    >>> for v in vs:
    ...     (f, n) = dump.write(-v)
    ...     tmp += [dump.read(n)]
    ...     if f is not None:
    ...         os.close(f)
    ...         os.remove(n)
    >>> for v in tmp:
    ...     print(v.__class__)
    <class 'fipy.variables.cellVariable.CellVariable'>
    <class 'fipy.variables.faceVariable.FaceVariable'>
    <class 'fipy.variables.variable.Variable'>
    >>> print(tmp[0].allclose(-2.))
    True
    >>> print(tmp[1].allclose(-3.))
    True
    >>> print(tmp[2].allclose(-4.))
    True
    """

    class unOp(operatorClass):
        def _calcValue_(self):
            return self.op(self.var[0].value)

        @property
        def unit(self):
            assert(hasattr(self, "_unit") == True)
            if self._unit is None:
                try:
                    var = self._varProxy
                    return self._extractUnit(self.op(var[0]))
                except:
                    return self._extractUnit(self._calcValue())
            else:
                return self._unit

    return unOp

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




from __future__ import unicode_literals
from builtins import object
__all__ = ["Preconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class Preconditioner(object):
    """
    Base preconditioner class

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    def __init__(self):
        """
        Create a `Preconditioner` object.
        """
        if self.__class__ is Preconditioner:
            raise NotImplementedError("can't instantiate abstract base class")

    def _applyToMatrix(self, matrix):
        """
        Returns the function used for Pysparse
        preconditioning.
        """
        raise NotImplementedError



from __future__ import unicode_literals
__all__ = []

from fipy.tests.lateImportTest import _LateImportTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportTestSuite(testModuleNames = (),
                                base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



from __future__ import division
from __future__ import unicode_literals
from builtins import range
__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.tools.numerix import random
from fipy.tools.dimensions.physicalField import PhysicalField
from fipy.tools import serialComm

from fipy.meshes.mesh2D import Mesh2D
from fipy.meshes import Grid2D

__all__ = ["SkewedGrid2D"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class SkewedGrid2D(Mesh2D):
    """
    Creates a 2D grid mesh with horizontal faces numbered first and then
    vertical faces.  The points are skewed by a random amount (between `rand`
    and `-rand`) in the X and Y directions.

    .. note:: This `Mesh` only operates in serial
    """
    def __init__(self, dx = 1., dy = 1., nx = None, ny = 1, rand = 0, *args, **kwargs):
        self.args = {
            'dx': dx,
            'dy': dy,
            'nx': nx,
            'ny': ny,
            'rand': rand
        }

        self.nx = nx
        self.ny = ny

        self.dx = PhysicalField(value = dx)
        scale = PhysicalField(value = 1, unit = self.dx.unit)
        self.dx /= scale

        self.dy = PhysicalField(value = dy)
        if self.dy.unit.isDimensionless():
            self.dy = dy
        else:
            self.dy /= scale

        self.grid = Grid2D(nx=nx, ny=ny, dx=dx, dy=dy, communicator=serialComm)

        self.numberOfVertices = self.grid._numberOfVertices

        vertices = self.grid.vertexCoords

        changedVertices = numerix.zeros(vertices.shape, 'd')

        for i in range(len(vertices[0])):
            if((i % (nx+1)) != 0 and (i % (nx+1)) != nx and (i // nx + 1) != 0 and (i // nx + 1) != ny):
                changedVertices[0, i] = vertices[0, i] + (rand * ((random.random() * 2) - 1))
                changedVertices[1, i] = vertices[1, i] + (rand * ((random.random() * 2) - 1))
            else:
                changedVertices[0, i] = vertices[0, i]
                changedVertices[1, i] = vertices[1, i]


        faces = self.grid.faceVertexIDs

        cells = self.grid.cellFaceIDs

        Mesh2D.__init__(self, changedVertices, faces, cells, communicator=serialComm, *args, **kwargs)

        self.scale = scale

    @property
    def physicalShape(self):
        """Return physical dimensions of `Grid2D`.
        """
        return PhysicalField(value = (self.nx * self.dx * self.scale, self.ny * self.dy * self.scale))

    @property
    def _meshSpacing(self):
        return numerix.array((self.dx, self.dy))[..., numerix.newaxis]

    @property
    def shape(self):
        return (self.nx, self.ny)



from __future__ import unicode_literals
__all__ = []

from fipy.variables.cellToFaceVariable import _CellToFaceVariable
from fipy.tools import numerix
from fipy.tools import inline

class _ArithmeticCellToFaceVariable(_CellToFaceVariable):
    if inline.doInline:
        def _calcValue_(self, alpha, id1, id2):
            val = self._array.copy()

            inline._runIterateElementInline("""
                int ID1 = ITEM(id1, i, NULL);
                int ID2 = ITEM(id2, i, NULL);
                double cell1 = ITEM(var, ID1, vec);
                double cell2 = ITEM(var, ID2, vec);
                ITEM(val, i, vec) = (cell2 - cell1) * ITEM(alpha, i, NULL) + cell1;
            """,
            var = self.var.numericValue,
            val = val,
            alpha = alpha,
            id1 = id1, id2 = id2,
            shape=numerix.array(numerix.shape(val)),
            ni = self.mesh.numberOfFaces)

            return self._makeValue(value = val)
    else:
        def _calcValue_(self, alpha, id1, id2):
            cell1 = numerix.take(self.var, id1, axis=-1)
            cell2 = numerix.take(self.var, id2, axis=-1)
            return (cell2 - cell1) * alpha + cell1



from __future__ import unicode_literals
from builtins import object
import copy

__all__ = ["AggregationAMGPreconditioner",
           "ClassicalAMGPreconditioner",
           "CGPreconditioner",
           "BiCGStabPreconditioner",
           "FGMRESPreconditioner",
           "BlockJacobiPreconditioner",
           "MultiColorDILUPreconditioner",
           "PolynomialPreconditioner",
           "MultiColorGSPreconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class Preconditioner(object):
    def __init__(self, preconditioner_type, **kwargs):
        self.config_dict = {
            "solver": preconditioner_type,
            "max_iters": 1
        }
        self.config_dict.update(kwargs)
    def __call__(self, **kwargs):
        """
        Parameters
        ----------
        **kwargs
            Other AMGX solver options
        """
        self.config_dict.update(kwargs)
        return copy.copy(self.config_dict)

AggregationAMGPreconditioner = Preconditioner("AMG", algorithm="AGGREGATION", selector="SIZE_2")
ClassicalAMGPreconditioner = Preconditioner("AMG")
CGPreconditioner = Preconditioner("PCG")
BiCGStabPreconditioner = Preconditioner("PCIBCGSTAB")
FGMRESPreconditioner = Preconditioner("FGMRES")
BlockJacobiPreconditioner = Preconditioner("BLOCK_JACOBI")
MultiColorDILUPreconditioner = Preconditioner("MULTICOLOR_DILU")
PolynomialPreconditioner = Preconditioner("POLYNOMIAL")
MultiColorGSPreconditioner = Preconditioner("MULTICOLOR_GS")



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools.dimensions.physicalField import PhysicalField

from fipy.tools import numerix

class _ModPhysicalField(PhysicalField):

    def mod(argument):
        return numerix.fmod(argument + 3. * numerix.pi, 2. * numerix.pi) - numerix.pi
    mod = staticmethod(mod)

    def __sub__(self, other):
        if isinstance(other, _ModPhysicalField):
            return self.__class__(value= self.mod(self.inRadians() - other.inRadians()), unit="rad")
        else:
            return self._sum(other, sign2 = lambda b: -b)

    def __rsub__(self, other):
        if isinstance(other, _ModPhysicalField):
            return self.__class__(value = self.mod(argument=other.inRadians() - self.inRadians()), unit="rad")
        else:
            return self._sum(other, sign1 = lambda a: -a)

    def ravel(self):
        return self.value.ravel()



from __future__ import division
from __future__ import unicode_literals
from builtins import object
from builtins import str
from future.utils import string_types
__docformat__ = 'restructuredtext'

import os

from fipy.tools.dimensions import physicalField
from fipy.tools import numerix
from fipy.tools import parser
from fipy.tools import inline

__all__ = ["Variable"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class Variable(object):
    """
    Lazily evaluated quantity with units.

    Using a :class:`~fipy.variables.variable.Variable` in a mathematical expression will create an
    automatic dependency :class:`~fipy.variables.variable.Variable`, e.g.,

    >>> a = Variable(value=3)
    >>> b = 4 * a
    >>> b
    (Variable(value=array(3)) * 4)
    >>> b()
    12

    Changes to the value of a :class:`~fipy.variables.variable.Variable` will automatically trigger
    changes in any dependent :class:`~fipy.variables.variable.Variable` objects

    >>> a.setValue(5)
    >>> b
    (Variable(value=array(5)) * 4)
    >>> print(b())
    20
    """

    _cacheAlways = (os.getenv("FIPY_CACHE") is not None) or False
    if parser.parse("--no-cache", action="store_true"):
        _cacheAlways = False
    if parser.parse("--cache", action="store_true"):
        _cacheAlways = True

    _cacheNever = False

    def __new__(cls, *args, **kwds):
        return object.__new__(cls)

    def __init__(self, value=0., unit=None, array=None, name='', cached=1):
        """
        Create a `Variable`.

            >>> Variable(value=3)
            Variable(value=array(3))
            >>> Variable(value=3, unit="m")
            Variable(value=PhysicalField(3,'m'))
            >>> Variable(value=3, unit="m", array=numerix.zeros((3, 2), 'l'))
            Variable(value=PhysicalField(array([[3, 3],
                   [3, 3],
                   [3, 3]]),'m'))

        Parameters
        ----------
        value : int or float or `array_like`
        unit : str or ~fipy.tools.dimensions.physicalField.PhysicalUnit
            The physical units of the variable
        array : ndarray, optional
            The storage array for the `Variable`
        name : str
            The user-readable name of the `Variable`
        cached : bool
            whether to cache or always recalculate the value
        """

        self.requiredVariables = []
        self.subscribedVariables = []

        if isinstance(value, Variable):
            value = value.value
            if hasattr(value, 'copy'):
                value = value.copy()
            unit = None

        self._setValueInternal(value=value, unit=unit, array=array)

        self._name = name

        self._cached = cached

        self.stale = 1
        self._markFresh()

##    __array_priority__ and __array_wrap__ are required to override
##    the default behavior of numpy. If a numpy array and a Variable
##    are in a binary operation and numpy is first, then numpy will,
##    by default, try and do everything it can to get a a raw numpy
##    array out of Variable. __array_wrap__ seems to have been
##    introduced into masked array to fix this issue. __array_wrap__ is
##    called after the operation is done so it could hurt efficiency badly.
##    Something else needs to be done to stop the initial evaluation.

    __array_priority__ = 100.0

    def __array_wrap__(self, arr, context=None):
        """
        Required to prevent numpy not calling the reverse binary operations.
        Both the following tests are examples ufuncs.

        >>> print(type(numerix.array([1.0, 2.0]) * Variable([1.0, 2.0]))) # doctest: +ELLIPSIS
        <class 'fipy.variables.binaryOperatorVariable...binOp'>

        >>> from scipy.special import gamma as Gamma # doctest: +SCIPY
        >>> print(type(Gamma(Variable([1.0, 2.0])))) # doctest: +SCIPY +ELLIPSIS
        <class 'fipy.variables.unaryOperatorVariable...unOp'>
        """
        result = arr

        if context is not None:
            from fipy.variables.constant import _Constant
            (func, args, _) = context
            def __makeVariable(v):
                if not isinstance(v, Variable):
                    v = _Constant(v)
                return v
            args = [__makeVariable(arg) for arg in args]

            cannotInline = ["expi", "logical_and", "logical_or", "logical_not", "logical_xor", "sign",
                            "conjugate", "dot", "allclose", "allequal"]
            if len(args) == 1:
                result = args[0]._UnaryOperatorVariable(op=func, opShape=arr.shape, canInline=func.__name__ not in cannotInline)
            elif len(args) == 2:
                result = args[0]._BinaryOperatorVariable(op=func, other=args[1], opShape=arr.shape, canInline=func.__name__ not in cannotInline)
            else:
                result = NotImplemented

        return result

    def __array__(self, t=None):
        """
        Attempt to convert the `Variable` to a numerix `array` object

            >>> v = Variable(value=[2, 3])
            >>> print(numerix.array(v))
            [2 3]

        A dimensional `Variable` will convert to the numeric value in its base units

            >>> v = Variable(value=[2, 3], unit="mm")
            >>> numerix.array(v)
            array([ 0.002,  0.003])
        """

        return numerix.array(self.value, t)

##    def _get_array_interface(self):
##        return self._array.__array_interface__

##    def _set_array_interface(self, value):
##        self._array.__array_interface__ = value

##    def _del_array_interface(self):
##        del self._array.__array_interface__

##    __array_interface__ = property(_get_array_interface,
##                                   _set_array_interface,
##                                   _del_array_interface,
##                                   "the '__array_inteface__'")

    def copy(self):
        """
        Make an duplicate of the `Variable`

            >>> a = Variable(value=3)
            >>> b = a.copy()
            >>> b
            Variable(value=array(3))

        The duplicate will not reflect changes made to the original

            >>> a.setValue(5)
            >>> b
            Variable(value=array(3))

        Check that this works for arrays.

            >>> a = Variable(value=numerix.array((0, 1, 2)))
            >>> b = a.copy()
            >>> b
            Variable(value=array([0, 1, 2]))
            >>> a[1] = 3
            >>> b
            Variable(value=array([0, 1, 2]))

        """
        return self._getArithmeticBaseClass()(value=self)

    @property
    def _unitAsOne(self):
        unit = self.unit
        if unit is physicalField._unity:
            return 1.
        else:
            return physicalField.PhysicalField(value=1, unit=unit)

    def _extractUnit(self, value):
        if isinstance(value, physicalField.PhysicalField):
            return value.unit
        else:
            return physicalField._unity

    def _getUnit(self):
        """
        Return the unit object of `self`.

            >>> Variable(value="1 m").unit
            <PhysicalUnit m>
        """
        return self._extractUnit(self.value)

    def _setUnit(self, unit):
        """
        Change the unit object of `self` to `unit`

            >>> a = Variable(value="1 m")
            >>> a.unit = "m**2/s"
            >>> print(a)
            1.0 m**2/s
        """
        if self._value is None:
            self.value

        if isinstance(self._value, physicalField.PhysicalField):
            self._value.unit = unit
        else:
            self._value = physicalField.PhysicalField(value=self._value, unit=unit)

    unit = property(_getUnit, _setUnit)

    def inBaseUnits(self):
        """
        Return the value of the `Variable` with all units reduced to
        their base SI elements.

            >>> e = Variable(value="2.7 Hartree*Nav")
            >>> print(e.inBaseUnits().allclose("7088849.01085 kg*m**2/s**2/mol"))
            1
        """
        value = self.value
        if isinstance(value, physicalField.PhysicalField):
            return value.inBaseUnits()
        else:
            return value

    def inUnitsOf(self, *units):
        """
        Returns one or more `Variable` objects that express the same
        physical quantity in different units.  The units are specified by
        strings containing their names.  The units must be compatible with
        the unit of the object.  If one unit is specified, the return value
        is a single `Variable`.

        >>> freeze = Variable('0 degC')
        >>> print(freeze.inUnitsOf('degF').allclose("32.0 degF"))
        1

        If several units are specified, the return value is a tuple of
        `Variable` instances with with one element per unit such that
        the sum of all quantities in the tuple equals the the original
        quantity and all the values except for the last one are integers.
        This is used to convert to irregular unit systems like
        hour/minute/second.  The original object will not be changed.

        >>> t = Variable(value=314159., unit='s')
        >>> from builtins import zip
        >>> print(numerix.allclose([e.allclose(v) for (e, v) in zip(t.inUnitsOf('d', 'h', 'min', 's'),
        ...                                                         ['3.0 d', '15.0 h', '15.0 min', '59.0 s'])],
        ...                        True))
        1
        """
        value = self.value
        if isinstance(value, physicalField.PhysicalField):
            return value.inUnitsOf(*units)
        else:
            return value

##     def __getitem__(self, index):
##         """
##         "Evaluate" the `Variable` and return the specified element
##
##            >>> ## a = Variable(value=((3.,4.),(5.,6.)), unit="m") + "4 m"
##            >>> ## print a[1,1]
##             10.0 m
##
##         It is an error to slice a `Variable` whose `value` is not sliceable
##
##            >>> ## Variable(value=3)[2]
##             Traceback (most recent call last):
##                   ...
##             IndexError: 0-d arrays can't be indexed
##
##         """
##         return (self.value)[index]

    def _getName(self):
        return self._name

    def _setName(self, name):
        self._name = name

    name = property(_getName, _setName)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        if hasattr(self, 'name') and len(self.name) > 0:
            return self.name
        else:
            s = self.__class__.__name__ + '('
            s += 'value=' + repr(self.value)
            s += ')'
            return s

    def _getCIndexString(self, shape):
        r"""
        Test for inline issue. `(1, ni)` shapes were not handled correctly.

        >>> from fipy import *
        >>> mesh = Tri2D(dx=1., dy=1., nx=1, ny=1)
        >>> diffCoeff = FaceVariable(mesh = mesh, value = 1.0)
        >>> normals = FaceVariable(mesh=mesh, rank=1, value=mesh._orientedFaceNormals)
        >>> normalsNthCoeff = diffCoeff[numerix.newaxis] * normals

        First variable value access does not provoke inlining.

        >>> value = normalsNthCoeff.value
        >>> print((normalsNthCoeff == value).all())
        True

        """

        dimensions = len(shape)
        if dimensions == 1:
            return '[i]'
        elif dimensions == 2:
            if shape[-1] == 1:
                return '[j]'
            elif shape[0] == 1:
                return '[i]'
            else:
                return '[i + j * ni]'
        elif dimensions == 3:
            if shape[-1] == 1:
                if shape[-2] == 1:
                    return '[k]'
                else:
                    return '[j + k * nj]'
            elif shape[-2] == 1:
                return '[i + k * ni * nj]'
            else:
                return '[i + j * ni + k * ni * nj]'


    def _getCstring(self, argDict={}, id="", freshen=None):
         """
         Generate the string and dictionary to be used in inline
             >>> from future.utils import text_to_native_str as ttns

             >>> ttns((Variable((1)))._getCstring(argDict={}))
             'var'

             >>> ttns((Variable((1, 2, 3, 4)))._getCstring(argDict={}))
             'var[i]'

             >>> ttns((Variable(((1, 2), (3, 4))))._getCstring(argDict={}))
             'var[i + j * ni]'

             >>> ttns(Variable((((1, 2), (3, 4)), ((5, 6), (7, 8))))._getCstring(argDict={}))
             'var[i + j * ni + k * ni * nj]'

             >>> ttns((Variable(1) * Variable((1, 2, 3)))._getCstring(argDict={}))
             '(var0 * var1[i])'

         freshen is ignored

         Testing when a cell variable multiplies an array that has a
         shape, but has only one element. This works regularly,
         but fails when inlining.

         >>> from fipy import *
         >>> m = Grid1D(nx=3)
         >>> x = m.cellCenters[0]
         >>> tmp = m.cellCenters[0] * numerix.array(((0.,), (1.,)))[1]
         >>> print(numerix.allclose(tmp, x))
         True
         >>> print(numerix.allclose(tmp, x))
         True

         """

         identifier = 'var%s' % (id)

         v = self.value

         if type(v) not in (type(numerix.array(1)),):
             varray = numerix.array(v)
         else:
             varray = v

         if len(varray.shape) == 0:
             if varray.dtype in (numerix.array(1).dtype,):
                 argDict[identifier] = int(varray)
             elif varray.dtype in (numerix.array(1.).dtype,):
                 argDict[identifier] = float(varray)
             else:
                 argDict[identifier] = varray
         else:
             argDict[identifier] = varray

         try:
             shape = self.opShape
         except AttributeError:
             shape = self.shape

         if len(shape) == 0:
             return identifier
         elif len(shape) == 1 and shape[0] == 1:
             return identifier + '[0]'
         else:
             return identifier + self._getCIndexString(shape)

    def tostring(self, max_line_width=75, precision=8, suppress_small=False, separator=' '):
        return numerix.tostring(self.value,
                                max_line_width=max_line_width,
                                precision=precision,
                                suppress_small=suppress_small,
                                separator=separator)

    def __setitem__(self, index, value):
        if self._value is None:
            self._getValue()
        self._value[index] = value
        self._markFresh()

    def itemset(self, value):
        if self._value is None:
            self._getValue()
        self._value.itemset(value)
        self._markFresh()

    def put(self, indices, value):
        if self._value is None:
            self._getValue()
        numerix.put(self._value, indices, value)
        self._markFresh()

    def __call__(self):
        """
        "Evaluate" the `Variable` and return its value

            >>> a = Variable(value=3)
            >>> print(a())
            3
            >>> b = a + 4
            >>> b
            (Variable(value=array(3)) + 4)
            >>> b()
            7
        """
        return self.value

    def _getValue(self):
        """
        "Evaluate" the `Variable` and return its value (longhand)

            >>> a = Variable(value=3)
            >>> print(a.value)
            3
            >>> b = a + 4
            >>> b
            (Variable(value=array(3)) + 4)
            >>> b.value
            7

        """

        if self.stale or not self._isCached() or self._value is None:
            value = self._calcValue()
            if self._isCached():
                self._setValueInternal(value=value)
            else:
                self._setValueInternal(value=None)
            self._markFresh()
        else:
            value = self._value

        if len(self.constraints) > 0:
            value = value.copy()
            for constraint in self.constraints:
                if constraint.where is None:
                    value[:] = constraint.value
                else:
                    mask = constraint.where
                    if not hasattr(mask, 'dtype') or mask.dtype != bool:
                        mask = numerix.array(mask, dtype=bool)

                    if 0 not in value.shape:
                        try:
                            value[..., mask] = constraint.value
                        except:
                            value[..., mask] = numerix.array(constraint.value)[..., mask]

        return value

    def _setValueProperty(self, newVal):
        """Since `self.setValue` contains optional, named parameters, we will
        punt the property's set method off to that."""
        self.setValue(newVal)

    value = property(_getValue, _setValueProperty)

    @property
    def constraints(self):
        if not hasattr(self, "_constraints"):
            self._constraints = []
        return self._constraints

    def constrain(self, value, where=None):
        """
        Constrain the `Variable` to have a `value` at an index or mask location specified by `where`.

        >>> v = Variable((0, 1, 2, 3))
        >>> v.constrain(2, numerix.array((True, False, False, False)))
        >>> print(v)
        [2 1 2 3]
        >>> v[:] = 10
        >>> print(v)
        [ 2 10 10 10]
        >>> v.constrain(5, numerix.array((False, False, True, False)))
        >>> print(v)
        [ 2 10  5 10]
        >>> v[:] = 6
        >>> print(v)
        [2 6 5 6]
        >>> v.constrain(8)
        >>> print(v)
        [8 8 8 8]
        >>> v[:] = 10
        >>> print(v)
        [8 8 8 8]
        >>> del v.constraints[2]
        >>> print(v)
        [ 2 10  5 10]

        >>> from fipy.variables.cellVariable import CellVariable
        >>> from fipy.meshes import Grid2D
        >>> m = Grid2D(nx=2, ny=2)
        >>> x, y = m.cellCenters
        >>> v = CellVariable(mesh=m, rank=1, value=(x, y))
        >>> v.constrain(((0.,), (-1.,)), where=m.facesLeft)
        >>> print(v.faceValue)
        [[ 0.5  1.5  0.5  1.5  0.5  1.5  0.   1.   1.5  0.   1.   1.5]
         [ 0.5  0.5  1.   1.   1.5  1.5 -1.   0.5  0.5 -1.   1.5  1.5]]

        Parameters
        ----------
        value : float or `array_like`
            The value of the constraint
        where : `array_like` of :obj:`bool`
            The constraint mask or index specifying the location of the
            constraint
        """

        from fipy.boundaryConditions.constraint import Constraint
        if not isinstance(value, Constraint):
            value = Constraint(value=value, where=where)

        if not hasattr(self, "_constraints"):
            self._constraints = []
        self._constraints.append(value)
        self._requires(value.value)
        self._markStale()

    def release(self, constraint):
        """Remove `constraint` from `self`

        >>> v = Variable((0, 1, 2, 3))
        >>> v.constrain(2, numerix.array((True, False, False, False)))
        >>> v[:] = 10
        >>> from fipy.boundaryConditions.constraint import Constraint
        >>> c1 = Constraint(5, numerix.array((False, False, True, False)))
        >>> v.constrain(c1)
        >>> v[:] = 6
        >>> v.constrain(8)
        >>> v[:] = 10
        >>> del v.constraints[2]
        >>> v.release(constraint=c1)
        >>> print(v)
        [ 2 10 10 10]
        """
        self.constraints.remove(constraint)

    def _isCached(self):
        return self._cacheAlways or (self._cached and not self._cacheNever)

    def cacheMe(self, recursive=False):
        self._cached = True
        if recursive:
            for var in self.requiredVariables:
                var.cacheMe(recursive=True)

    def dontCacheMe(self, recursive=False):
        self._cached = False
        if recursive:
            for var in self.requiredVariables:
                var.dontCacheMe(recursive=False)

    def _setValueInternal(self, value, unit=None, array=None):
        self._value = self._makeValue(value=value, unit=unit, array=array)

    def _makeValue(self, value, unit=None, array=None):

        ## --inline code often returns spurious results with non-contiguous
        ## arrays. A test case was put in _execInline(). The best fix turned out to
        ## be here.

        if (inline.doInline
            and hasattr(value, 'iscontiguous') and not value.iscontiguous()):
            value = value.copy()

        if isinstance(value, Variable):
            value = value.value

        PF = physicalField.PhysicalField

        if not isinstance(value, PF):

            if getattr(self, '_value', None) is not None:
                v = self._value
                if isinstance(v, PF):
                    v = self._value.value
                if type(value) in (type(1), type(1.)):
                    if isinstance(v, type(numerix.array(1))):
                        if v.shape is not ():
##                        if len(v) > 1:
                            value = numerix.resize(value, v.shape).astype(v.dtype)

            if (unit is not None
                or isinstance(value, string_types)
                or isinstance(value, tuple)
                or isinstance(value, list)):
                value = PF(value=value, unit=unit, array=array)
            elif array is not None:
                array[:] = value
                value = array
            elif type(value) not in (type(None), type(numerix.array(1)), type(numerix.MA.array(1))):
                value = numerix.array(value)
##                 # numerix does strange things with really large integers.
##                 # Even though Python knows how to do arithmetic with them,
##                 # Numeric converts them to 'O' objects that it then doesn't understand.
##                 if value.typecode() == 'O':
##                     value = numerix.array(float(value))

        if isinstance(value, PF) and value.unit.isDimensionless():
            value = value.numericValue

        return value

    def setValue(self, value, unit=None, where=None):
        """
        Set the value of the Variable. Can take a masked array.

            >>> a = Variable((1, 2, 3))
            >>> a.setValue(5, where=(1, 0, 1))
            >>> print(a)
            [5 2 5]

            >>> b = Variable((4, 5, 6))
            >>> a.setValue(b, where=(1, 0, 1))
            >>> print(a)
            [4 2 6]
            >>> print(b)
            [4 5 6]
            >>> a.value = 3
            >>> print(a)
            [3 3 3]

            >>> b = numerix.array((3, 4, 5))
            >>> a.value = b
            >>> a[:] = 1
            >>> print(b)
            [3 4 5]

            >>> a.setValue((4, 5, 6), where=(1, 0)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ....
            ValueError: shape mismatch: objects cannot be broadcast to a single shape

        """
        if where is not None:
            tmp = numerix.empty(numerix.getShape(where), self.getsctype())
            tmp[:] = value
            tmp = numerix.where(where, tmp, self.value)
        else:
            if hasattr(value, 'copy'):
                tmp = value.copy()
            else:
                tmp = value

        value = self._makeValue(value=tmp, unit=unit, array=None)

        if numerix.getShape(self._value) == ():
            self._value.itemset(value)
        else:
            self._value[:] = value

        self._markFresh()

    def _setNumericValue(self, value):
        if isinstance(self._value, physicalField.PhysicalField):
            self._value.value = value
        else:
            self._value = value

    @property
    def _array(self):
        if isinstance(self._value, physicalField.PhysicalField):
            return self._value._array
        else:
            return self._value

    @property
    def numericValue(self):
        value = self.value
        if isinstance(value, physicalField.PhysicalField):
            return value.numericValue
        else:
            return value

    def _getShape(self):
        """
        Tuple of array dimensions.

        >>> Variable(value=3).shape
        ()
        >>> numerix.allequal(Variable(value=(3,)).shape, (1,))
        True
        >>> numerix.allequal(Variable(value=(3, 4)).shape, (2,))
        True

        >>> Variable(value="3 m").shape
        ()
        >>> numerix.allequal(Variable(value=(3,), unit="m").shape, (1,))
        True
        >>> numerix.allequal(Variable(value=(3, 4), unit="m").shape, (2,))
        True
        """
        if self._value is not None:
            return numerix.getShape(self._value)
        else:
            return ()

    shape = property(_getShape)

    def getsctype(self, default=None):
        """

        Returns the Numpy `sctype` of the underlying array.

            >>> Variable(1).getsctype() == numerix.NUMERIX.obj2sctype(numerix.array(1))
            True
            >>> Variable(1.).getsctype() == numerix.NUMERIX.obj2sctype(numerix.array(1.))
            True
            >>> Variable((1, 1.)).getsctype() == numerix.NUMERIX.obj2sctype(numerix.array((1., 1.)))
            True

        """

        if not hasattr(self, 'typecode'):
            self.typecode = numerix.obj2sctype(rep=self.numericValue, default=default)

        return self.typecode

    @property
    def itemsize(self):
        return self.value.itemsize

    def _calcValue(self):
        return self._value

    def _calcValueNoInline(self):
        raise NotImplementedError

    def _calcValueInline(self):
        raise NotImplementedError

    def _getSubscribedVariables(self):
        self._subscribedVariables = [sub for sub in self._subscribedVariables if sub() is not None]

        return self._subscribedVariables

    def _setSubscribedVariables(self, sVars):
        self._subscribedVariables = sVars

    subscribedVariables = property(_getSubscribedVariables,
                                   _setSubscribedVariables)

    def __markStale(self):
        for subscriber in self.subscribedVariables:
            if subscriber() is not None:
                ## Even though getSubscribedVariables() strips out dead
                ## references, subscriber() might still be dead due to the
                ## vagaries of garbage collection and the possibility that
                ## later subscribedVariables were removed, changing the
                ## dependencies of this subscriber.
                ## See <https://github.com/usnistgov/fipy/issues/103> for more explanation.
                subscriber()._markStale()

    def _markFresh(self):
        self.stale = 0
        self.__markStale()

    def _markStale(self):
        if not self.stale:
            self.stale = 1
            self.__markStale()

    def _requires(self, var):
        if isinstance(var, Variable):
            self.requiredVariables.append(var)
            var._requiredBy(self)
        else:
            from fipy.variables.constant import _Constant
            var = _Constant(value=var)
        self._markStale()
        return var

    def _requiredBy(self, var):
        assert isinstance(var, Variable)

        # we retain a weak reference to avoid a memory leak
        # due to circular references between the subscriber
        # and the subscribee
        import weakref
        self.subscribedVariables.append(weakref.ref(var))

    @property
    def _variableClass(self):
        return Variable

    def _execInline(self, comment=None):
        """
        Gets the stack from `_getCstring()` which calls `_getRepresentation()`

            >>> from future.utils import text_to_native_str as ttns

            >>> ttns((Variable((1, 2, 3, 4)) * Variable((5, 6, 7, 8)))._getCstring())
            '(var0[i] * var1[i])'
            >>> ttns((Variable(((1, 2), (3, 4))) * Variable(((5, 6), (7, 8))))._getCstring())
            '(var0[i + j * ni] * var1[i + j * ni])'
            >>> ttns((Variable((1, 2)) * Variable((5, 6)) * Variable((7, 8)))._getCstring())
            '((var00[i] * var01[i]) * var1[i])'

        The following test was implemented due to a problem with
        contiguous arrays.  The `mesh.cellCenters[1]` command
        introduces a non-contiguous array into the `Variable` and this
        causes the inline routine to return senseless results.

            >>> from fipy import Grid2D, CellVariable
            >>> mesh = Grid2D(dx=1., dy=1., nx=2, ny=2)
            >>> var = CellVariable(mesh=mesh, value=0.)
            >>> Y =  mesh.cellCenters[1]
            >>> var.value = (Y + 1.0)
            >>> print(var - Y)
            [ 1.  1.  1.  1.]
        """

        from fipy.tools import inline
        argDict = {}
        string = self._getCstring(argDict=argDict, freshen=True) + ';'

        try:
            shape = self.opShape
        except AttributeError:
            shape = self.shape

        dimensions = len(shape)

        if dimensions == 0:
            string = 'result[0] = ' + string
            dim = ()
        else:
            string = 'result' + self._getCIndexString(shape) + ' = ' + string
            ni = self.opShape[-1]
            argDict['ni'] = ni
            if dimensions == 1:
                dim = (ni)
            else:
                nj = self.opShape[-2]
                argDict['nj'] = nj
                if dimensions == 2:
                    dim =(nj, ni)
                elif dimensions == 3:
                    nk = self.opShape[-3]
                    dim = (nk, nj, ni)
                    argDict['nk'] = nk
                else:
                    raise DimensionError('Impossible Dimensions')

        ## Following section makes sure that the result array has a
        ## valid typecode. If self._value is None then a typecode is
        ## assigned to the Variable by running the calculation without
        ## inlining. The non-inlined result is thus used the first
        ## time through.


        if self._value is None and not hasattr(self, 'typecode'):
            self.canInline = False
            argDict['result'] = self.value
            self.canInline = True
            self.typecode = numerix.obj2sctype(argDict['result'])
        else:
            if self._value is None:
                if self.getsctype() == numerix.bool_:
                    argDict['result'] = numerix.empty(dim, numerix.int8)
                else:
                    argDict['result'] = numerix.empty(dim, self.getsctype())
            else:
                argDict['result'] = self._value

            resultShape = argDict['result'].shape

            if resultShape == ():
                argDict['result'] = numerix.reshape(argDict['result'], (1,))

            inline._runInline(string, converters=None, comment=comment, **argDict)

            if resultShape == ():
                argDict['result'] = numerix.reshape(argDict['result'], resultShape)

            if self.getsctype() == numerix.bool_:
                argDict['result'] = numerix.asarray(argDict['result'], dtype=self.getsctype())

        return argDict['result']

    def _broadcastShape(self, other):
        ignore, ignore, broadcastshape = numerix._broadcastShapes(self.shape, numerix.getShape(other))

        return broadcastshape

##         selfshape = self.shape
##         othershape = other.shape
##
##         if len(selfshape) > len(othershape):
##             othershape = (1,) * (len(selfshape) - len(othershape)) + othershape
##         elif len(selfshape) < len(othershape):
##             selfshape = (1,) * (len(othershape) - len(selfshape)) + selfshape
##
##         if numerix.logical_and.reduce([(s == o or s == 1 or o == 1) for s,o in zip(selfshape, othershape)]):
##             return tuple([max(s,o) for s,o in zip(selfshape, othershape)])
##         else:
##             return None

    def _getArithmeticBaseClass(self, other=None):
        """
        Given `self` and `other`, return the desired base class for an operation
        result.
        """
        if other is None:
            return Variable

        if self._broadcastShape(other) is not None:
            # If self and other have the same base class, result has that base class.
            # If self derives from other, result has self's base class.
            # If other derives from self, result has other's base class.
            # If self and other don't have a common base, we don't know how to combine them.
            from fipy.variables.constant import _Constant
            if isinstance(self, other._getArithmeticBaseClass()) or isinstance(other, _Constant):
                return self._getArithmeticBaseClass()
            else:
                return None
        else:
            # If self and other have un-broadcastable shapes, we don't know how to combine them.
            return None

    def _OperatorVariableClass(self, baseClass=None):
        from fipy.variables import operatorVariable

        baseClass = baseClass or self._variableClass
        return operatorVariable._OperatorVariableClass(baseClass=baseClass)

    def _UnaryOperatorVariable(self, op, operatorClass=None, opShape=None, canInline=True, unit=None,
                               valueMattersForUnit=False):
        """
        Check that unit works for `unOp`

            >>> (-Variable(value="1 m")).unit
            <PhysicalUnit m>

        Parameters
        ----------
        op : func
            The operator function to apply (takes one argument for `self`)
        operatorClass : class
            The `Variable` class that the binary operator should inherit
            from
        opShape : :obj:`tuple` of :obj:`int`
            The shape that should result from the operation
        valueMattersForUnit : bool
            Whether value of `self` should be used when determining unit,
            e.g., ???
        """
        operatorClass = operatorClass or self._OperatorVariableClass()
        from fipy.variables import unaryOperatorVariable
        unOp = unaryOperatorVariable._UnaryOperatorVariable(operatorClass)

        # If the caller has not specified a shape for the result, determine the
        # shape from the base class or from the inputs
        if opShape is None:
            opShape = self.shape

        if opShape is None:
            return NotImplemented

        if not self.unit.isDimensionless():
            canInline = False

        return unOp(op=op, var=[self], opShape=opShape, canInline=canInline, unit=unit,
                    inlineComment=inline._operatorVariableComment(canInline=canInline),
                    valueMattersForUnit=[valueMattersForUnit])

    def _shapeClassAndOther(self, opShape, operatorClass, other):
        """
        Determine the shape of the result, the base class of the result, and (if
        necessary) a modified form of `other` that is suitable for the
        operation.
        """
        # If the caller has not specified a base class for the binop,
        # check if the member Variables know what type of Variable should
        # result from the operation.
        baseClass = operatorClass or self._getArithmeticBaseClass(other)

        # If the caller has not specified a shape for the result, determine the
        # shape from the base class or from the inputs
        if opShape is None:
            opShape = self._broadcastShape(other)

        return (opShape, baseClass, other)

    def _BinaryOperatorVariable(self, op, other, operatorClass=None, opShape=None, canInline=True, unit=None,
                                value0mattersForUnit=False, value1mattersForUnit=False):
        """
        Parameters
        ----------
        op : func
            The operator function to apply (takes two arguments for `self` and `other`)
        other : ~fipy.variables.variable.Variable
            The quantity to be operated with
        operatorClass : class
            The `Variable` class that the binary operator should inherit
            from
        opShape : :obj:`tuple` of :obj:`int`
            The shape that should result from the operation
        value0MattersForUnit : bool
            Whether value of `self` should be used when determining unit,
            e.g., `__rpow__`
        value1MattersForUnit : bool
            Whether value of `self` should be used when determining unit,
            e.g., `__pow__`
        """
        if not isinstance(other, Variable):
            from fipy.variables.constant import _Constant
            other = _Constant(value=other)

        opShape, baseClass, other = self._shapeClassAndOther(opShape, operatorClass, other)

        if opShape is None or baseClass is None:
            return NotImplemented

        for v in [self, other]:
            if not v.unit.isDimensionless() or len(v.shape) > 3:
                canInline = False

        # obtain a general operator class with the desired base class
        operatorClass = operatorClass or self._OperatorVariableClass(baseClass)
        from fipy.variables import binaryOperatorVariable
        binOp = binaryOperatorVariable._BinaryOperatorVariable(operatorClass)

        return binOp(op=op, var=[self, other], opShape=opShape, canInline=canInline, unit=unit,
                     inlineComment=inline._operatorVariableComment(canInline=canInline),
                     valueMattersForUnit=[value0mattersForUnit, value1mattersForUnit])

    def __add__(self, other):
        from fipy.terms.term import Term
        if isinstance(other, Term):
            return other + self
        else:
            return self._BinaryOperatorVariable(lambda a, b: a+b, other)

    __radd__ = __add__

    def __sub__(self, other):
        from fipy.terms.term import Term
        if isinstance(other, Term):
            return -other + self
        else:
            return self._BinaryOperatorVariable(lambda a, b: a-b, other)

    def __rsub__(self, other):
        return self._BinaryOperatorVariable(lambda a, b: b-a, other)

    def __mul__(self, other):
        from fipy.terms.term import Term
        if isinstance(other, Term):
            return other * self
        else:
            return self._BinaryOperatorVariable(lambda a, b: a*b, other)

    __rmul__ = __mul__

    def __mod__(self, other):
        return self._BinaryOperatorVariable(lambda a, b: numerix.fmod(a, b), other)

    def __pow__(self, other):
        """return self**other, or self raised to power other

        >>> print(Variable(1, "mol/l")**3)
        1.0 mol**3/l**3
        >>> print((Variable(1, "mol/l")**3).unit)
        <PhysicalUnit mol**3/l**3>
        """
        return self._BinaryOperatorVariable(lambda a, b: pow(a, b), other, value1mattersForUnit=True)

    def __rpow__(self, other):
        return self._BinaryOperatorVariable(lambda a, b: pow(b, a), other, value0mattersForUnit=True)

    def __truediv__(self, other):
        return self._BinaryOperatorVariable(lambda a, b: a / b, other)

    __div__ = __truediv__

    def __rtruediv__(self, other):
        return self._BinaryOperatorVariable(lambda a, b: b / a, other)

    __rdiv__ = __rtruediv__

    def __neg__(self):
        return self._UnaryOperatorVariable(lambda a: -a)

    def __pos__(self):
        return self

    def __abs__(self):
        """

        Following test it to fix a bug with C inline string using
        `abs()` instead of `fabs()`

            >>> print(abs(Variable(2.3) - Variable(1.2)))
            1.1


        Check representation works with different versions of numpy

            >>> print(repr(abs(Variable(2.3))))
            numerix.fabs(Variable(value=array(2.3)))

        """
        return self._UnaryOperatorVariable(lambda a: numerix.fabs(a))

    def __invert__(self):
        """
        Returns logical "not" of the `Variable`

            >>> a = Variable(value=True)
            >>> print(~a)
            False
        """
        return self._UnaryOperatorVariable(lambda a: ~a)

    def __lt__(self, other):
        """
        Test if a `Variable` is less than another quantity

            >>> a = Variable(value=3)
            >>> b = (a < 4)
            >>> b
            (Variable(value=array(3)) < 4)
            >>> b()
            1
            >>> a.value = 4
            >>> print(b())
            0
            >>> print(1000000000000000000 * Variable(1) < 1.)
            0
            >>> print(1000 * Variable(1) < 1.)
            0


        Python automatically reverses the arguments when necessary

            >>> 4 > Variable(value=3)
            (Variable(value=array(3)) < 4)
        """
        return self._BinaryOperatorVariable(lambda a, b: a<b, other)

    def __le__(self, other):
        """
        Test if a `Variable` is less than or equal to another quantity

            >>> a = Variable(value=3)
            >>> b = (a <= 4)
            >>> b
            (Variable(value=array(3)) <= 4)
            >>> b()
            1
            >>> a.value = 4
            >>> print(b())
            1
            >>> a.value = 5
            >>> print(b())
            0
        """
        return self._BinaryOperatorVariable(lambda a, b: a<=b, other)

    def __eq__(self, other):
        """
        Test if a `Variable` is equal to another quantity

            >>> a = Variable(value=3)
            >>> b = (a == 4)
            >>> b
            (Variable(value=array(3)) == 4)
            >>> b()
            0
        """
        return self._BinaryOperatorVariable(lambda a, b: a==b, other)

    __hash__ = object.__hash__

    def __ne__(self, other):
        """
        Test if a `Variable` is not equal to another quantity

            >>> a = Variable(value=3)
            >>> b = (a != 4)
            >>> b
            (Variable(value=array(3)) != 4)
            >>> b()
            1
        """
        return self._BinaryOperatorVariable(lambda a, b: a!=b, other)

    def __gt__(self, other):
        """
        Test if a `Variable` is greater than another quantity

            >>> a = Variable(value=3)
            >>> b = (a > 4)
            >>> b
            (Variable(value=array(3)) > 4)
            >>> print(b())
            0
            >>> a.value = 5
            >>> print(b())
            1
        """
        return self._BinaryOperatorVariable(lambda a, b: a>b, other)

    def __ge__(self, other):
        """
        Test if a `Variable` is greater than or equal to another quantity

            >>> a = Variable(value=3)
            >>> b = (a >= 4)
            >>> b
            (Variable(value=array(3)) >= 4)
            >>> b()
            0
            >>> a.value = 4
            >>> print(b())
            1
            >>> a.value = 5
            >>> print(b())
            1
        """
        return self._BinaryOperatorVariable(lambda a, b: a>=b, other)

    def __and__(self, other):
        """
        This test case has been added due to a weird bug that was appearing.

        >>> a = Variable(value=(0, 0, 1, 1))
        >>> b = Variable(value=(0, 1, 0, 1))
        >>> print(numerix.equal((a == 0) & (b == 1), [False,  True, False, False]).all())
        True
        >>> print(a & b)
        [0 0 0 1]
        >>> from fipy.meshes import Grid1D
        >>> mesh = Grid1D(nx=4)
        >>> from fipy.variables.cellVariable import CellVariable
        >>> a = CellVariable(value=(0, 0, 1, 1), mesh=mesh)
        >>> b = CellVariable(value=(0, 1, 0, 1), mesh=mesh)
        >>> print(numerix.allequal((a == 0) & (b == 1), [False,  True, False, False]))
        True
        >>> print(a & b)
        [0 0 0 1]
        """
        return self._BinaryOperatorVariable(lambda a, b: a & b, other, canInline=False)

    def __or__(self, other):
        """
        This test case has been added due to a weird bug that was appearing.

        >>> a = Variable(value=(0, 0, 1, 1))
        >>> b = Variable(value=(0, 1, 0, 1))
        >>> print(numerix.equal((a == 0) | (b == 1), [True,  True, False, True]).all())
        True
        >>> print(a | b)
        [0 1 1 1]
        >>> from fipy.meshes import Grid1D
        >>> mesh = Grid1D(nx=4)
        >>> from fipy.variables.cellVariable import CellVariable
        >>> a = CellVariable(value=(0, 0, 1, 1), mesh=mesh)
        >>> b = CellVariable(value=(0, 1, 0, 1), mesh=mesh)
        >>> print(numerix.allequal((a == 0) | (b == 1), [True,  True, False, True]))
        True
        >>> print(a | b)
        [0 1 1 1]
        """
        return self._BinaryOperatorVariable(lambda a, b: a | b, other, canInline=False)

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def __bool__(self):
        """
            >>> print(bool(Variable(value=0)))
            0
            >>> print(bool(Variable(value=(0, 0, 1, 1))))
            Traceback (most recent call last):
                ...
            ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()
        """
        return bool(self.value)

    __nonzero__ = __bool__

    def any(self, axis=None):
        """
            >>> print(Variable(value=0).any())
            0
            >>> print(Variable(value=(0, 0, 1, 1)).any())
            1
        """
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._UnaryOperatorVariable(lambda a: a.any(axis=axis),
                                           operatorClass=operatorClass,
                                           opShape=(),
                                           canInline=False)

    def all(self, axis=None):
        """
            >>> print(Variable(value=(0, 0, 1, 1)).all())
            0
            >>> print(Variable(value=(1, 1, 1, 1)).all())
            1
        """
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._UnaryOperatorVariable(lambda a: a.all(axis=axis),
                                           operatorClass=operatorClass,
                                           opShape=(),
                                           canInline=False)

    def dot(self, other, opShape=None, operatorClass=None, axis=0):
        if not isinstance(other, Variable):
            from fipy.variables.constant import _Constant
            other = _Constant(value=other)
        if opShape is None:
            opShape = self._broadcastShape(other)
        return self._BinaryOperatorVariable(lambda a, b: numerix.dot(a, b, axis=axis),
                                            other,
                                            opShape=opShape[:axis]+opShape[axis+1:],
                                            operatorClass=operatorClass,
                                            canInline=False)

    def ravel(self):
        return self.value.ravel()

    def _axisClass(self, axis):
        return self._OperatorVariableClass()

    def _axisOperator(self, opname, op, axis=None):
        if not hasattr(self, opname):
            setattr(self, opname, {})

        opdict = getattr(self, opname)
        if axis not in opdict:
            if axis is None:
                opShape = ()
            else:
                opShape=self.shape[:axis] + self.shape[axis+1:]

            opdict[axis] = self._UnaryOperatorVariable(op,
                                                       operatorClass=self._axisClass(axis=axis),
                                                       opShape=opShape,
                                                       canInline=False)

        return opdict[axis]

    def sum(self, axis=None):
        return self._axisOperator(opname="sumVar",
                                  op=lambda a: numerix.sum(a, axis=axis),
                                  axis=axis)

    def max(self, axis=None):
        return self._axisOperator(opname="maxVar",
                                  op=lambda a: a.max(axis=axis),
                                  axis=axis)

    def min(self, axis=None):
        return self._axisOperator(opname="minVar",
                                  op=lambda a: a.min(axis=axis),
                                  axis=axis)

    def std(self, axis=None, **kwargs):
        return self._axisOperator(opname="stdVar",
                                  op=lambda a: numerix.std(a, **kwargs),
                                  axis=axis)

    def _getitemClass(self, index):
        return self._OperatorVariableClass()

    def __getitem__(self, index):
        """
        "Evaluate" the `Variable` and return the specified element

            >>> a = Variable(value=((3., 4.), (5., 6.)), unit="m") + "4 m"
            >>> print(a[1, 1])
            10.0 m

        It is an error to slice a `Variable` whose `value` is not sliceable

            >>> Variable(value=3)[2]
            Traceback (most recent call last):
                  ...
            IndexError: 0-d arrays can't be indexed

        """
        return self._UnaryOperatorVariable(lambda a: a[index],
                                           operatorClass=self._getitemClass(index=index),
                                           opShape=numerix._indexShape(index=index, arrayShape=self.shape),
                                           unit=self.unit,
                                           canInline=False)

    def take(self, ids, axis=0):
        return numerix.take(self.value, ids, axis)

    def allclose(self, other, rtol=1.e-5, atol=1.e-8):
        """
           >>> var = Variable((1, 1))
           >>> print(var.allclose((1, 1)))
           1
           >>> print(var.allclose((1,)))
           1

        The following test is to check that the system does not run
        out of memory.

           >>> from fipy.tools import numerix
           >>> var = Variable(numerix.ones(10000))
           >>> print(var.allclose(numerix.zeros(10000, 'l')))
           False

        """
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._BinaryOperatorVariable(lambda a, b: numerix.allclose(a, b, atol=atol, rtol=rtol),
                                            other,
                                            operatorClass=operatorClass,
                                            opShape=(),
                                            canInline=False)

    def allequal(self, other):
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._BinaryOperatorVariable(lambda a, b: numerix.allequal(a, b),
                                            other,
                                            operatorClass=operatorClass,
                                            opShape=(),
                                            canInline=False)

    @property
    def mag(self):
        r"""The magnitude of the :class:`~fipy.variables.variable.Variable`,
        e.g., :math:`|\vec{\psi}| = \sqrt{\vec{\psi}\cdot\vec{\psi}}`."""
        if not hasattr(self, "_mag"):
            self._mag = numerix.sqrt(self.dot(self))

        return self._mag

    def __getstate__(self):
        """
        Used internally to collect the necessary information to ``pickle`` the
        `Variable` to persistent storage.
        """
        return {
            'value': self.value,
            'unit': self.unit,
            'array': None,
            'name': self.name,
        }

    def __setstate__(self, dict):
        """
        Used internally to create a new `Variable` from ``pickled``
        persistent storage.
        """

        import sys
        self._refcount = sys.getrefcount(self)

        self.__init__(**dict)

    def _test(self):
        r"""
        Inverse cosine of :math:`x`, :math:`\cos^{-1} x`

        >>> from fipy.tools.numerix import *

        >>> arccos(Variable(value=(0, 0.5, 1.0)))
        arccos(Variable(value=array([ 0. ,  0.5,  1. ])))

        .. attention::

           the next should really return radians, but doesn't

        >>> print(tostring(arccos(Variable(value=(0, 0.5, 1.0))), precision=3))
        [ 1.571  1.047  0.   ]

        Inverse hyperbolic cosine of :math:`x`, :math:`\cosh^{-1} x`

        >>> arccosh(Variable(value=(1, 2, 3)))
        arccosh(Variable(value=array([1, 2, 3])))
        >>> print(tostring(arccosh(Variable(value=(1, 2, 3))), precision=3))
        [ 0.     1.317  1.763]

        Inverse sine of :math:`x`, :math:`\sin^{-1} x`

        >>> arcsin(Variable(value=(0, 0.5, 1.0)))
        arcsin(Variable(value=array([ 0. ,  0.5,  1. ])))

        .. attention::

           the next should really return radians, but doesn't

        >>> print(tostring(arcsin(Variable(value=(0, 0.5, 1.0))), precision=3))
        [ 0.     0.524  1.571]

        Inverse hyperbolic sine of :math:`x`, :math:`\sinh^{-1} x`

        >>> arcsinh(Variable(value=(1, 2, 3)))
        arcsinh(Variable(value=array([1, 2, 3])))
        >>> print(tostring(arcsinh(Variable(value=(1, 2, 3))), precision=3))
        [ 0.881  1.444  1.818]

        Inverse tangent of :math:`x`, :math:`\tan^{-1} x`

        >>> arctan(Variable(value=(0, 0.5, 1.0)))
        arctan(Variable(value=array([ 0. ,  0.5,  1. ])))

        .. attention::

           the next should really return radians, but doesn't

        >>> print(tostring(arctan(Variable(value=(0, 0.5, 1.0))), precision=3))
        [ 0.     0.464  0.785]

        Inverse tangent of a ratio :math:`x/y`, :math:`\tan^{-1} \frac{x}{y}`

        >>> arctan2(Variable(value=(0, 1, 2)), 2)
        (arctan2(Variable(value=array([0, 1, 2])), 2))

        .. attention::

           the next should really return radians, but doesn't

        >>> print(tostring(arctan2(Variable(value=(0, 1, 2)), 2), precision=3))
        [ 0.     0.464  0.785]

        Inverse hyperbolic tangent of :math:`x`, :math:`\tanh^{-1} x`

        >>> arctanh(Variable(value=(0, 0.25, 0.5)))
        arctanh(Variable(value=array([ 0.  ,  0.25,  0.5 ])))
        >>> print(tostring(arctanh(Variable(value=(0, 0.25, 0.5))), precision=3))
        [ 0.     0.255  0.549]

        Cosine of :math:`x`, :math:`\cos x`

        >>> cos(Variable(value=(0, 2*pi/6, pi/2), unit="rad"))
        cos(Variable(value=PhysicalField(array([ 0.        ,  1.04719755,  1.57079633]),'rad')))
        >>> print(tostring(cos(Variable(value=(0, 2*pi/6, pi/2), unit="rad")), suppress_small=1))
        [ 1.   0.5  0. ]

        Hyperbolic cosine of :math:`x`, :math:`\cosh x`

        >>> cosh(Variable(value=(0, 1, 2)))
        cosh(Variable(value=array([0, 1, 2])))
        >>> print(tostring(cosh(Variable(value=(0, 1, 2))), precision=3))
        [ 1.     1.543  3.762]

        Tangent of :math:`x`, :math:`\tan x`

        >>> tan(Variable(value=(0, pi/3, 2*pi/3), unit="rad"))
        tan(Variable(value=PhysicalField(array([ 0.        ,  1.04719755,  2.0943951 ]),'rad')))
        >>> print(tostring(tan(Variable(value=(0, pi/3, 2*pi/3), unit="rad")), precision=3))
        [ 0.     1.732 -1.732]

        Hyperbolic tangent of :math:`x`, :math:`\tanh x`

        >>> tanh(Variable(value=(0, 1, 2)))
        tanh(Variable(value=array([0, 1, 2])))
        >>> print(tostring(tanh(Variable(value=(0, 1, 2))), precision=3))
        [ 0.     0.762  0.964]

        Base-10 logarithm of :math:`x`, :math:`\log_{10} x`

        >>> log10(Variable(value=(0.1, 1, 10)))
        log10(Variable(value=array([  0.1,   1. ,  10. ])))
        >>> print(log10(Variable(value=(0.1, 1, 10))))
        [-1.  0.  1.]

        Sine of :math:`x`, :math:`\sin x`

        >>> sin(Variable(value=(0, pi/6, pi/2), unit="rad"))
        sin(Variable(value=PhysicalField(array([ 0.        ,  0.52359878,  1.57079633]),'rad')))
        >>> print(sin(Variable(value=(0, pi/6, pi/2), unit="rad")))
        [ 0.   0.5  1. ]

        Hyperbolic sine of :math:`x`, :math:`\sinh x`

        >>> sinh(Variable(value=(0, 1, 2)))
        sinh(Variable(value=array([0, 1, 2])))
        >>> print(tostring(sinh(Variable(value=(0, 1, 2))), precision=3))
        [ 0.     1.175  3.627]

        Square root of :math:`x`, :math:`\sqrt{x}`

        >>> sqrt(Variable(value=(1, 2, 3), unit="m**2"))
        sqrt(Variable(value=PhysicalField(array([1, 2, 3]),'m**2')))
        >>> print(tostring(sqrt(Variable(value=(1, 2, 3), unit="m**2")), precision=3))
        [ 1.     1.414  1.732] m

        The largest integer :math:`\le x`, :math:`\lfloor x \rfloor`

        >>> floor(Variable(value=(-1.5, 2, 2.5), unit="m**2"))
        floor(Variable(value=PhysicalField(array([-1.5,  2. ,  2.5]),'m**2')))
        >>> print(floor(Variable(value=(-1.5, 2, 2.5), unit="m**2")))
        [-2.  2.  2.] m**2

        The largest integer :math:`\ge x`, :math:`\lceil x \rceil`

        >>> ceil(Variable(value=(-1.5, 2, 2.5), unit="m**2"))
        ceil(Variable(value=PhysicalField(array([-1.5,  2. ,  2.5]),'m**2')))
        >>> print(ceil(Variable(value=(-1.5, 2, 2.5), unit="m**2")))
        [-1.  2.  3.] m**2

        Natural logarithm of :math:`x`, :math:`\ln x \equiv \log_e x`

        >>> log(Variable(value=(0.1, 1, 10)))
        log(Variable(value=array([  0.1,   1. ,  10. ])))
        >>> print(tostring(log(Variable(value=(0.1, 1, 10))), precision=3))
        [-2.303  0.     2.303]

        Complex conjugate of :math:`z = x + i y`, :math:`z^\star = x - i y`

        >>> var = conjugate(Variable(value=(3 + 4j, -2j, 10), unit="ohm"))
        >>> print(var.unit)
        <PhysicalUnit ohm>
        >>> print(allclose(var.numericValue, (3 - 4j, 2j, 10)))
        1
        """
        pass


def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



"""setuptools command to prepare FiPy for release"""

from distutils.core import Command
import glob
import os
import shutil

from setuptools.sandbox import run_setup

__all__ = ["release"]


class release(Command):
    """Prepare FiPy for release

    Generates tarball and excutable Windows installer"""

    description = "Prepare the FiPy release artifacts"

    # List of option tuples: long name, short name (None if no short
    # name), and help string.
    user_options = [('unix', None, "create a tarball source distribution"),
                    ('windows', None, "create an executable installer for MS Windows"),
                    ('all', None, "create unix and Windows distributions"),
                   ]

    def initialize_options(self):
        self.unix = 0
        self.windows = 0
        self.all = 0

    def finalize_options(self):
        if self.all:
            self.unix = 1
            self.windows = 1

    def _remove_manifest(self):
        """Remove MANIFEST file

        probably no longer needed, MANIFEST was ancient history?"""

        try:
            os.remove("MANIFEST")
        except OSError as _:
            pass

    def _build_source_distribution(self, formats):
        """Create source distribution"""

        self._remove_manifest()
        run_setup("setup.py", ["sdist", "--formats=" + ",".join(formats)])

    def run(self):
        formats = []
        if self.unix:
            formats.append("gztar")
        if self.windows:
            formats.append("zip")
        self._build_source_distribution(formats)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.tools.numerix import random, sqrt
from fipy.variables.noiseVariable import NoiseVariable

__all__ = ["GaussianNoiseVariable"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class GaussianNoiseVariable(NoiseVariable):
    r"""

    Represents a normal (Gaussian) distribution of random numbers with
    mean :math:`\mu` and variance
    :math:`\langle \eta(\vec{r}) \eta(\vec{r}\,') \rangle = \sigma^2`,
    which has the probability distribution

    .. math::

       \frac{1}{\sigma\sqrt{2\pi}} \exp -\frac{(x - \mu)^2}{2\sigma^2}

    For example, the variance of thermal noise that is uncorrelated in space and
    time is often expressed as

    .. math::

       \left\langle
           \eta(\vec{r}, t) \eta(\vec{r}\,', t')
       \right\rangle =
       M k_B T \delta(\vec{r} - \vec{r}\,')\delta(t - t')

    which can be obtained with::

        sigmaSqrd = Mobility * kBoltzmann * Temperature / (mesh.cellVolumes * timeStep)
        GaussianNoiseVariable(mesh = mesh, variance = sigmaSqrd)

    .. note::

       If the time step will change as the simulation progresses, either through
       use of an adaptive iterator or by making manual changes at different
       stages, remember to declare `timeStep` as a `Variable` and to change its
       value with its `setValue()` method.

    >>> import sys
    >>> from fipy.tools.numerix import *

    >>> mean = 0.
    >>> variance = 4.

    Seed the random module for the sake of deterministic test results.

    >>> from fipy import numerix
    >>> numerix.random.seed(3)

    We generate noise on a non-uniform Cartesian mesh with cell dimensions of
    :math:`x^2` and :math:`y^3`.

    >>> from fipy.meshes import Grid2D
    >>> mesh = Grid2D(dx = arange(0.1, 5., 0.1)**2, dy = arange(0.1, 3., 0.1)**3)
    >>> from fipy.variables.cellVariable import CellVariable
    >>> volumes = CellVariable(mesh=mesh, value=mesh.cellVolumes)
    >>> noise = GaussianNoiseVariable(mesh = mesh, mean = mean,
    ...                               variance = variance / volumes)

    We histogram the root-volume-weighted noise distribution

    >>> from fipy.variables.histogramVariable import HistogramVariable
    >>> histogram = HistogramVariable(distribution = noise * sqrt(volumes),
    ...                               dx = 0.1, nx = 600, offset = -30)

    and compare to a Gaussian distribution

    >>> gauss = CellVariable(mesh = histogram.mesh)
    >>> x = histogram.mesh.cellCenters[0]
    >>> gauss.value = ((1/(sqrt(variance * 2 * pi))) * exp(-(x - mean)**2 / (2 * variance)))

    >>> if __name__ == '__main__':
    ...     from fipy.viewers import Viewer
    ...     viewer = Viewer(vars=noise,
    ...                     datamin=-5, datamax=5)
    ...     histoplot = Viewer(vars=(histogram, gauss))

    >>> from builtins import range
    >>> for i in range(10):
    ...     noise.scramble()
    ...     if __name__ == '__main__':
    ...         viewer.plot()
    ...         histoplot.plot()

    >>> print(abs(noise.faceGrad.divergence.cellVolumeAverage) < 5e-15)
    1

    Note that the noise exhibits larger amplitude in the small cells than in the large ones

    .. image:: fipy/variables/gaussian.*
      :scale: 25
      :align: center
      :alt: random values with a Gaussian distribution

    but that the root-volume-weighted histogram is Gaussian.

    .. image:: fipy/variables/gauss-histogram.*
      :scale: 25
      :align: center
      :alt: histogram of random values with a Gaussian distribution

    """
    def __init__(self, mesh, name = '', mean = 0., variance = 1., hasOld = 0):
        """
        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The mesh on which to define the noise.
        mean : float
            The mean of the noise distribution, :math:`\mu`.
        variance : float
            The variance of the noise distribution, :math:`\sigma^2`.
        """
        self.mean = mean
        self.variance = variance
        NoiseVariable.__init__(self, mesh = mesh, name = name, hasOld = hasOld)

    def parallelRandom(self):

        if hasattr(self.variance, 'globalValue'):
            variance = self.variance.globalValue
        else:
            variance = self.variance

        if self.mesh.communicator.procID == 0:
            return random.normal(self.mean, sqrt(variance),
                                 size = [self.mesh.globalNumberOfCells])
        else:
            return None

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.abstractDiffusionTerm import _AbstractDiffusionTerm
from fipy.tools import numerix

__all__ = ["DiffusionTermCorrection"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class DiffusionTermCorrection(_AbstractDiffusionTerm):

    def _getNormals(self, mesh):
        return mesh._faceCellToCellNormals

    def _treatMeshAsOrthogonal(self, mesh):
        return mesh._isOrthogonal



from distutils.core import Command
import os

__all__ = ["build_docs"]

class build_docs(Command):

    description = "build the FiPy documentation"

    # List of option tuples: long name, short name (None if no short
    # name), and help string.
    user_options = [('pdf', None, "compile the PDF variant of the documentation"),
                    ('html', None, "compile the HTML variant of the documentation"),
                    ('cathartic', None, "rewrite all the files (default is to only rewrite changed files)"),
                   ]

    def initialize_options (self):
        self.pdf = 0
        self.html = 0
        self.cathartic = 0

    def finalize_options (self):
        pass

    def run (self):
        import sphinx.cmd.build
        import sphinx.ext.apidoc
        
        sphinx_args = ['-P', '-n', '-c', 'documentation/', '.']
        apidoc_args = []
        
        if self.cathartic:
            sphinx_args = ['-a', '-E'] + sphinx_args
            apidoc_args = ['--force'] + apidoc_args
            
        sphinx.ext.apidoc.main(['--output-dir=fipy/generated', '--suffix=rst']
                    + apidoc_args + ['fipy'])
        sphinx.ext.apidoc.main(['--output-dir=documentation/tutorial/package/generated', '--suffix=rst']
                    + apidoc_args + ['documentation/tutorial/package'])

        if self.html:
            sphinx.cmd.build.main(['-b', 'redirecting_html'] + sphinx_args + ['documentation/_build/html/'])

        if self.pdf:
            try:
                sphinx.cmd.build.main(['-b', 'latex'] + sphinx_args + ['documentation/_build/latex/'])
            except SystemExit:
                pass
            
            outdir = os.path.join('documentation', '_build', 'latex')
            
            from docutils.core import publish_file

            for xtra in ("LICENSE", "DISCLAIMER"):
                publish_file(source_path="%s.rst" % xtra,
                             destination_path=os.path.join(outdir, "%s.tex" % xtra),
                             reader_name='standalone',
                             parser_name='restructuredtext',
                             writer_name='latex',
                             settings_overrides= {
                                 'template': 'documentation/_templates/empty.tex'
                             })

            savedir = os.getcwd()
            
            os.chdir(outdir)
                
            os.system("pdflatex fipy")
            os.system("pdflatex fipy")
            os.system("pdflatex fipy")
            os.system("makeindex -s python.ist fipy")
            os.system("makeindex -s python.ist modfipy")
            os.system("pdflatex fipy")
            os.system("pdflatex fipy")
                
            os.chdir(savedir)
